/*=======================================================
  Author:   R. Muth
  Version:  $Id: cmdline.c,v 1.41 1999/06/14 23:05:59 muth Exp $
  Comment:  This file contains the defintion of the
            structure "Global" and the command line
            parsing code
=======================================================*/

#include "global.h"
#include "functions.h"
#ifdef SYS_COFF
#include "system.coff.h"
#elif SYS_ELF
#include "system.elf.h"
#endif

#include <string.h>

/*=======================================================
Global Variables
=======================================================*/
STRUCT_BIN    theBinary;
STRUCT_SEC   aSec[CONFIG_NSEC];
STRUCT_FUN   aFun[CONFIG_NFUN];
STRUCT_BBL   aBbl[CONFIG_NBBL];
STRUCT_EDG   aEdg[CONFIG_NEDG];
STRUCT_INS   aIns[CONFIG_NINS];
STRUCT_STK   aStk[CONFIG_NSTK];
STRUCT_TAB   aTab[CONFIG_NTAB];

QUADWORD     uBbl1[CONFIG_NBBL]; /* multi purpose */
QUADWORD     uBbl2[CONFIG_NBBL]; /* multi purpose */
QUADWORD     uBbl3[CONFIG_NBBL]; /* multi purpose */

QUADWORD     uIns1[CONFIG_NINS];    /* multi purpose */

QUADWORD     uEdg1[CONFIG_NEDG];    /* multi purpose */
QUADWORD     uEdg2[CONFIG_NEDG];    /* multi purpose */

INS_EVAL     eInsA[CONFIG_NINS];    /* reg eval A */
INS_EVAL     eInsB[CONFIG_NINS];    /* reg eval B */
INS_EVAL     eInsC[CONFIG_NINS];    /* reg eval C */


SCHEDULEE    aSchedulee[SCHEDULEES_ALLOCATED];
SCHLOT       aSchlot[SCHLOTS_ALLOCATED];
CONSTRAINT   aConstraint[CONSTRAINTS_ALLOCATED];
CANDIDATE    aCandidate[CANDIDATES_ALLOCATED];
    

GLOBAL Global=
{
    /* Config */

    "Crit",        /* STRING    InlineMethod; */
    "Sensitive",   /* STRING    LivenessMethod; */

    " ",           /* STRING    PrintOptWhat; */
    "Postopt",     /* STRING    PrintOptWhen; */
    "",            /* STRING PrintOptVerbose */ 
    -1,            /* STRING    PrintOptNum; */
    "",            /* STRING    PrintOptFun; */
    "",            /* STRING    DisableOpt; */
    
    4,             /* NUMBER FunAlign; */
    4,             /* NUMBER LoopAlign; */
    0,             /* NUMBER BblAlign; */

    0,             /* BOOL   UseProfile; */
    "",            /* STRING ProfileWhat; */

    INVALID,       /* NUMBER InlineExpansion */
    
    2,             /* NUMBER Rounds; */
    2048,          /* NUMBER    IcacheSize; */

    INVALID,       /* NUMBER Count; */
    INVALID,       /* NUMBER CallCount; */

    "a.out",       /* STRING InFile; */
    "b.out",       /* STRING OutFile; */

    0,             /* UINT32 uSection; */
    0,             /* UINT32 uFunction; */
    0,             /* UINT32 uBblock; */
    0,             /* UINT32 uEdges; */
    0,             /* UINT32 uInstr; */
    0,             /* UINT32 uStack; */
    0,             /* UINT32 uTab; */
    0,             /* UINT32 uConstraints; */
    0,             /* UINT32 uCandidates; */
    0,             /* UINT32 uSchedulees; */
    0,             /* UINT32 uSchlots; */
    0,             /* NUMBER CONFIG_NVALUES */

    /* BINARY    Binary; */

};


/*=======================================================
@FUN
Print Alto usage information
@END
=======================================================*/

extern void Usage()
{
    Error("Usage: \n");
    exit(-1);
}

/*=======================================================
@FUN
Parse command line and set internal flags accordingly.
Allocate space for main datastructures (the size of which should
evetually be configurable by  commandline options)
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ParseCommandLine"

extern void ParseCommandLine(int argc, char *argv[])
{
    int i;
    
    while( EOF != (i = getopt(argc,argv,"i:o:r:f:b:l:hO:S:P:R:T:F:N:C:M:e:V:L:")) )
    {
        switch(i)
        {
        case 'f':
            CONFIG_FUN_ALIGN = atoi(optarg);
            break;
        case 'b':
            CONFIG_BBL_ALIGN = atoi(optarg);
            break;
        case 'l':
            CONFIG_LOOP_ALIGN = atoi(optarg);
            break;
        case 'P':
	    if (strcmp(optarg, "All") == 0)
            {
		CONFIG_PROFILE_WHAT = "Bbl+Edg+Call+Jump";
	    }
	    else
            {
		CONFIG_PROFILE_WHAT = optarg;
	    }
            break;
        case 'S':
            PRINT_OPT_WHAT = optarg;
            break;
        case 'T':
            PRINT_OPT_WHEN = optarg;
            break;
        case 'V':
            PRINT_OPT_VERBOSE = optarg;
            break;
        case 'F':
            PRINT_OPT_FUN = optarg;
            break;
        case 'O':
            DISABLED_OPT = optarg;
            break;
        case 'N':
            PRINT_OPT_NUM = strtol(optarg, (char **) NULL, 0);
            break;
        case 'o':
            CONFIG_OUTFILE = optarg;
            break;
        case 'C':
            CONFIG_ICACHE_SIZE = atoi(optarg);
            break;
        case 'M':
            INLINE_METHOD = optarg;
            break;
        case 'L':
            LIVENESS_METHOD = optarg;
            break;
        case 'e':
            INLINE_EXPANSION = atoi(optarg);
            break;
        case 'r':
            CONFIG_ROUNDS = atoi(optarg);
            break;
        case 'i':
            CONFIG_INFILE = optarg;            
            break;            
        case 'h':
        case '?':
        default:
            Usage();
            break;
        }
    }

    /*
    aSec = (STRUCT_SEC*) calloc(CONFIG_NSEC,sizeof(STRUCT_SEC));
    aFun = (STRUCT_FUN*) calloc(CONFIG_NFUN,sizeof(STRUCT_FUN));
    aBbl = (STRUCT_BBL*) calloc(CONFIG_NBBL,sizeof(STRUCT_BBL));
    aEdg = (STRUCT_EDG*) calloc(CONFIG_NEDG,sizeof(STRUCT_EDG));
    aIns = (STRUCT_INS*) calloc(CONFIG_NINS,sizeof(STRUCT_INS));
    aStk = (STRUCT_STK*) calloc(CONFIG_NSTK,sizeof(STRUCT_STK));
    aTab = (STRUCT_TAB*) calloc(CONFIG_NTAB,sizeof(STRUCT_TAB));
    */
}


/*=======================================================
=======================================================*/

