/*=======================================================
Author:   R. Muth
Version:  $Id: emul.c,v 1.14 1999/06/04 23:47:28 muth Exp $
Comment:  This file contains the instruction emulation stuff

=======================================================*/

#include "global.h"
#include "functions.h"
#include "regs.h"

#include "opcode.h"
#include "flags.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/mode.h>

/*=======================================================
=======================================================*/

typedef  INT64  (*EVAL_IOP)(INT64 arg1, INT64 arg2);
typedef  INT64  (*EVAL_ILD)(INT64 arg1);
typedef  double (*EVAL_FOP)(double arg1, double arg2);
typedef  double (*EVAL_FLD)(INT64 arg1);

#define PAL_IMB_OPCODE  0x00000086
#define RET_OPCODE      0x6bfa8001
#define NOP_OPCODE      0x2ffe0000
#define ILLEGAL         0x00000000
/*=======================================================
=======================================================*/

#define BITS_OPC(n)   ((n>>26) & 0x3f)
#define MAKE_OPC(n)   ((n & 0x3f) << 26)
#define BITS_FFUN(n)  ((n>>5)  & 0x7ff)
#define MAKE_FFUN(n)  ((n & 0x7ff) << 5)
#define BITS_IFUN(n)  ((n>>5)  & 0x7f) 
#define MAKE_IFUN(n)  ((n & 0x7f) << 5)
#define BITS_MFUN(n)  (n & 0xffff)
#define MAKE_MFUN(n)  (n & 0xffff)
#define BITS_JFUN(n)  ((n>>14) & 0x3)
#define MAKE_JFUN(n)  ((n & 0x3) << 14)
#define BITS_ISLIT(n) ((n>>12) & 0x1)
#define MAKE_LIT      (0x1<<12)
#define BITS_IMMED(n) ((n>>13) & 0xff)
#define MAKE_IMMED(n) ((n & 0xff)  << 13 )
#define BITS_REGC(n)  (n & 0x1f)
#define MAKE_REGC(n)  (n & 0x1f)
#define BITS_REGB(n)  ((n>>16) & 0x1f)
#define MAKE_REGB(n)  ((n & 0x1f) << 16)
#define BITS_REGA(n)  ((n>>21) & 0x1f)
#define MAKE_REGA(n)  ((n & 0x1f) << 21)
#define BITS_BDISP(n) (n & 0x1fffff)
#define MAKE_BDISP(n) (n & 0x1fffff)
#define BITS_BSIGN(n) (n & 0x100000)                       
#define BITS_MDISP(n) (n & 0xffff)
#define MAKE_MDISP(n) (n & 0xffff)
#define BITS_MSIGN(n) (n & 0x8000)                                              
#define BITS_PAL(n)   (n & 0x3ffffff)
#define MAKE_PAL(n)   (n & 0x3ffffff)
#define BITS_HINT(n)  (n & 0x3fff)
#define MAKE_HINT(n)  (n & 0x3fff)

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "Disassemble"

static int find(int Index, int FunCode)
{
    while(FunCode > InstructionTable[Index].FunCode)
        Index++;
    
    return Index;
}


extern void Disassemble(INT32 opcode, STRUCT_INS *pi)
{
    int Index = BITS_OPC(opcode);

#if 1
    CASSERT(InstructionTable[I_beg10].Code == I_beg10,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg11].Code == I_beg11,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg12].Code == I_beg12,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg13].Code == I_beg13,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg14].Code == I_beg14,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg15].Code == I_beg15,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg16].Code == I_beg16,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg17].Code == I_beg17,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg18].Code == I_beg18,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg1a].Code == I_beg1a,("rerun genenum.perl\n"));
    CASSERT(InstructionTable[I_beg1c].Code == I_beg1c,("rerun genenum.perl\n"));
#endif    
                                            
    switch( InstructionTable[Index].Type )
    {
    case IT_X10:
        Index = find(1+I_beg10,BITS_IFUN(opcode));
        break;
    case IT_X11:
        Index = find(1+I_beg11,BITS_IFUN(opcode));
        break;
    case IT_X12:
        Index = find(1+I_beg12,BITS_IFUN(opcode));
        break;
    case IT_X13:
        Index = find(1+I_beg13,BITS_IFUN(opcode));
        break;
    case IT_X14:
        Index = find(1+I_beg14,BITS_FFUN(opcode));
        break;
    case IT_X15:
        Index = find(1+I_beg15,BITS_FFUN(opcode));
        break;
    case IT_X16:
        Index = find(1+I_beg16,BITS_FFUN(opcode));
        break;
    case IT_X17:
        Index = find(1+I_beg17,BITS_FFUN(opcode));
        break;
    case IT_X18:
        Index = find(1+I_beg18,BITS_MFUN(opcode));
        break;
    case IT_X1A:
        Index = find(1+I_beg1a,BITS_JFUN(opcode));
        break;                
    case IT_X1C:
        Index = find(1+I_beg1c,BITS_IFUN(opcode));
        break;
    default:    /* other case fall through */
        break;
    }
    
    pi->Code = InstructionTable[Index].Code;
    CASSERT( pi->Code == Index,("bad instruction table\n"));

    switch( InstructionTable[Index].Type )
    {
    case IT_ICM:
    case IT_IOP:
    case IT_ICP:
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
    case IT_EXT:
    case IT_LOG:
    case IT_SFT:
        pi->RegA = BITS_REGA(opcode);
        pi->RegC = BITS_REGC(opcode);
        if( BITS_ISLIT(opcode) )
        {
            pi->RegB = REG_LIT;
            pi->Data = BITS_IMMED(opcode);
        }
        else
        {
            pi->RegB = BITS_REGB(opcode);
            pi->Data = 0;
        }
        break;

    case IT_F2R:
    case IT_R2F:
    case IT_FCM:
    case IT_FCP:
    case IT_FA4:
    case IT_FA8:
    case IT_FS4:
    case IT_FS8:
    case IT_FM4:
    case IT_FM8:
    case IT_FD4:
    case IT_FD8:
    case IT_FCV:
    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
        pi->RegA = BITS_REGA(opcode) + REG_FLOAT;
        pi->RegB = BITS_REGB(opcode) + REG_FLOAT;
        pi->RegC = BITS_REGC(opcode) + REG_FLOAT;
        pi->Data = 0;
        break;

    case IT_MSC:
    case IT_MEM:
    case IT_LDA:
    case IT_ILD:
    case IT_LLD:
    case IT_ULD:
        pi->RegA = REG_ZERO;
        pi->RegC = BITS_REGA(opcode);                 /* tricky */
        pi->RegB = BITS_REGB(opcode);
        pi->Data = BITS_MDISP(opcode);
        if( BITS_MSIGN(pi->Data) )
            pi->Data |= -1 ^ 0xffff;
        break;

    case IT_FLD:
        pi->RegA = REG_ZERO;
        pi->RegC = BITS_REGA(opcode) + REG_FLOAT;     /* tricky */
        pi->RegB = BITS_REGB(opcode);
        pi->Data = BITS_MDISP(opcode);
        if( BITS_MSIGN(pi->Data) )
            pi->Data |= -1 ^ 0xffff;
        break;

    case IT_IST:
    case IT_UST:
        pi->RegC = REG_ZERO;
        pi->RegA = BITS_REGA(opcode);                
        pi->RegB = BITS_REGB(opcode);
        pi->Data = BITS_MDISP(opcode);
        if( BITS_MSIGN(pi->Data) )
            pi->Data |= -1 ^ 0xffff;
        break;

    case IT_CST:
        pi->RegC = BITS_REGA(opcode);                /* special */
        pi->RegA = BITS_REGA(opcode);                
        pi->RegB = BITS_REGB(opcode);
        pi->Data = BITS_MDISP(opcode);
        if( BITS_MSIGN(pi->Data) )
            pi->Data |= -1 ^ 0xffff;
        break;
        
    case IT_FST:
        pi->RegC = REG_ZERO;
        pi->RegA = BITS_REGA(opcode) + REG_FLOAT;     
        pi->RegB = BITS_REGB(opcode);
        pi->Data = BITS_MDISP(opcode);
        if( BITS_MSIGN(pi->Data) )
            pi->Data |= -1 ^ 0xffff;            
        break;        

    case IT_F2I:
        pi->RegB = REG_ZERO;
        pi->RegA = BITS_REGA(opcode) + REG_FLOAT;     
        pi->RegC = BITS_REGC(opcode);
        pi->Data = 0;
        break;

    case IT_I2F:
        pi->RegB = REG_ZERO;
        pi->RegA = BITS_REGA(opcode);
        pi->RegC = BITS_REGC(opcode) + REG_FLOAT;
        pi->Data = 0;
        break;


    case IT_BRA:
        pi->RegA = REG_ZERO;
        pi->RegB = REG_ZERO;
        pi->RegC = BITS_REGA(opcode);             /* tricky */
        pi->Data = BITS_BDISP(opcode);
        if( BITS_BSIGN(pi->Data) )
            pi->Data |= -1 ^ 0x1fffff;
        /* sign extend */
        break;
        
    case IT_IBR:
        pi->RegA = BITS_REGA(opcode);
        pi->RegB = REG_ZERO;
        pi->RegC = REG_ZERO;
                  
        pi->Data = BITS_BDISP(opcode);
        if( BITS_BSIGN(pi->Data) )
            pi->Data |= -1 ^ 0x1fffff;
        /* sign extend */
        break;
        
    case IT_FBR:
        pi->RegA = BITS_REGA(opcode) + REG_FLOAT;
        pi->RegB = REG_ZERO;
        pi->RegC = REG_ZERO;
        pi->Data = BITS_BDISP(opcode);
        if( BITS_BSIGN(pi->Data) )
            pi->Data |= -1 ^ 0x1fffff;
        break;
        
    case IT_JMP:
        pi->RegA = REG_ZERO;
        pi->RegB = BITS_REGB(opcode);
        pi->RegC = BITS_REGA(opcode);             /* tricky */
#if 0
        pi->Data = -1;
#else
        pi->Data = BITS_HINT(opcode);
#endif
        break;
       
    case IT_PAL:
        pi->RegA = REG_ZERO;
        pi->RegB = REG_ZERO;
        pi->RegC = REG_ZERO;
        pi->Data = BITS_PAL(opcode);
        break;
        
    default:
        FATAL(("illegal instruction type %08x %02x\n",opcode,BITS_IFUN(opcode)));
        break;
    }
}


/*=======================================================
=======================================================*/
#define AINS_CODE_STR(ii) InstructionTable[AINS_CODE(ii)].Name

extern void InstructionPrint(FILE *fp,INDEX ii, INT64 address)
{
    switch( AINS_TYPE(ii) )
    {
    case IT_ICM:
    case IT_IOP:
    case IT_ICP:
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
    case IT_EXT:
    case IT_LOG:
    case IT_SFT:

    case IT_F2R:
    case IT_R2F:
    case IT_FCM:
    case IT_FCP:
    case IT_FA4:
    case IT_FA8:
    case IT_FS4:
    case IT_FS8:
    case IT_FM4:
    case IT_FM8:
    case IT_FD4:
    case IT_FD8:
    case IT_FCV:
    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
        if( AINS_REG_B(ii) != REG_LIT )
            fprintf(fp, "%-8s $%-2d,$%-2d,$%-2d",
                    AINS_CODE_STR(ii),AINS_REG_A(ii),AINS_REG_B(ii),AINS_REG_C(ii));
        else
            fprintf(fp, "%-8s $%-2d,0x%02x,$%-2d",
                    AINS_CODE_STR(ii),AINS_REG_A(ii),AINS_DATA(ii),AINS_REG_C(ii));
        break;

    case IT_FLD:
    case IT_ILD:
    case IT_LDA:
    case IT_ULD:
    case IT_LLD:
    case IT_MEM:
    case IT_MSC:
        fprintf(fp, "%-8s $%-2d,0x%04x($%d)",
                AINS_CODE_STR(ii),AINS_REG_C(ii),AINS_DATA(ii)&0xffff,AINS_REG_B(ii));
        break;
        
    case IT_IST:
    case IT_FST:
    case IT_UST:
    case IT_CST:
        fprintf(fp, "%-8s $%-2d,0x%04x($%d)",
                AINS_CODE_STR(ii),AINS_REG_A(ii),AINS_DATA(ii)&0xffff,AINS_REG_B(ii));
        break;
        
    case IT_I2F:
    case IT_F2I:
        fprintf(fp, "%-8s $%-2d,$%-2d",
                AINS_CODE_STR(ii),AINS_REG_A(ii),AINS_REG_C(ii));
        break;

    case IT_BRA:
        if( VALID(AINS_BBL(ii)) )
        {
            if( AINS_CODE(ii) == I_bsr ) 
            {
                INDEX target = AEDG_TAIL(ABBL_EDG_CALL(AINS_BBL(ii)));
                fprintf(fp, "%-8s $%-2d,bbl %d fun %d [%6s]",
                        AINS_CODE_STR(ii),AINS_REG_C(ii),target,ABBL_FUN(target),AFUN_NAME(ABBL_FUN(target)));
            }
            else
            {
                INDEX target = AEDG_TAIL(ABBL_EDG_UNCOND(AINS_BBL(ii)));
                fprintf(fp, "%-8s $%-2d,bbl %d",
                        AINS_CODE_STR(ii),AINS_REG_C(ii),target);
            }
        }
        else 
        {
            fprintf(fp, "%-8s $%-2d,0x%06lx",
                    AINS_CODE_STR(ii),AINS_REG_C(ii),(AINS_DATA(ii) + 1) * 4 + address);
        }
        break;
    case IT_IBR:
    case IT_FBR:
        if( VALID(AINS_BBL(ii)) )
        {
            INDEX tbbl = AEDG_TAIL(ABBL_EDG_TRUE(AINS_BBL(ii)));
            INDEX fbbl = AEDG_TAIL(ABBL_EDG_FALSE(AINS_BBL(ii)));
            
            fprintf(fp, "%-8s $%-2d,t %d f %d",
                    AINS_CODE_STR(ii),AINS_REG_A(ii),tbbl,fbbl);
        }
        else
        {
            fprintf(fp, "%-8s $%-2d,0x%06lx",
                    AINS_CODE_STR(ii),AINS_REG_A(ii),(AINS_DATA(ii) + 1) * 4 + address);
        }
        
        break;
        
    case IT_JMP:
        fprintf(fp, "%-8s $%-2d ($%d),0x%06x",
                AINS_CODE_STR(ii),AINS_REG_C(ii),AINS_REG_B(ii),AINS_DATA(ii));
        break;
        
    case IT_PAL:
        fprintf(fp, "%-8s 0x%08x",
                AINS_CODE_STR(ii),AINS_DATA(ii));
        break;        

    default:
        ERROR(("illegal instruction type %02x",AINS_TYPE(ii)));
        break;
    }
    
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "Assemble"

extern INT32 Assemble(STRUCT_INS *pi)
{
    switch( InstructionTable[pi->Code].Type )
    {
    case IT_ICM:
    case IT_IOP:
    case IT_ICP:
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
    case IT_EXT:
    case IT_LOG:
    case IT_SFT:

                if( pi->RegB == REG_LIT )
            return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
                MAKE_IFUN(InstructionTable[pi->Code].FunCode) +
                MAKE_REGA( pi->RegA ) +
                MAKE_LIT +
                MAKE_IMMED(pi->Data) +
                MAKE_REGC( pi->RegC );
        else
            return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
                MAKE_IFUN(InstructionTable[pi->Code].FunCode) +
                MAKE_REGA( pi->RegA ) +
                MAKE_REGB( pi->RegB ) +
                MAKE_REGC( pi->RegC );

    case IT_FCM:
    case IT_FCP:
    case IT_FA4:
    case IT_FA8:
    case IT_FS4:
    case IT_FS8:
    case IT_FM4:
    case IT_FM8:
    case IT_FD4:
    case IT_FD8:
    case IT_FCV:
    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
    case IT_F2R:
    case IT_R2F:
        
        return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
            MAKE_FFUN(InstructionTable[pi->Code].FunCode) +
            MAKE_REGA( pi->RegA ) +
            MAKE_REGB( pi->RegB ) +
            MAKE_REGC( pi->RegC );

    case IT_ILD:
    case IT_FLD:
    case IT_LDA:
    case IT_ULD:
    case IT_LLD:
    case IT_MEM:
    case IT_MSC:
        
    return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
                MAKE_REGA( pi->RegC ) + /* tricky */
                MAKE_REGB( pi->RegB ) +
                MAKE_MDISP( pi->Data );

    case IT_CST:
        CASSERT( pi->RegA == pi->RegC,("store conditional assumption violated\n"));
    case IT_IST:
    case IT_UST:
    case IT_FST:
        return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
            MAKE_REGA( pi->RegA ) +
            MAKE_REGB( pi->RegB ) +
            MAKE_MDISP( pi->Data );

    case IT_BRA:
        return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
            MAKE_REGA( pi->RegC ) +     /* tricky */
            MAKE_BDISP( pi->Data );
        
    case IT_IBR:
    case IT_FBR:
        return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
            MAKE_REGA( pi->RegA ) +
            MAKE_BDISP( pi->Data );
        
    case IT_JMP:
        return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
            MAKE_JFUN(InstructionTable[pi->Code].FunCode) +
            MAKE_REGA( pi->RegC ) +     /* tricky */
            MAKE_REGB( pi->RegB ) +
            MAKE_HINT( pi->Data );

    case IT_PAL:
        return MAKE_OPC(InstructionTable[pi->Code].OpCode) +
            MAKE_PAL( pi->Data );

    default:
        ERROR(("illegal instruction: %s\n",InstructionTable[pi->Code].Name));
        return 0;
    }
    
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/

static char data[128 * 1024] = {1};

static char *emul;


#undef __PROC__
#define __PROC__ "EmulInitialize"

extern void EmulInitialize()
{
    INT64 ps;
    INT32 opcode;
    int i;
    
    ps = sysconf(_SC_PAGE_SIZE);
    CASSERT( (ps & (ps-1)) == 0,("bad page size %ld\n",ps));
    CASSERT( ps < 64 *1024,("page too big\n"));
    
#if 0
    emul = (char *) malloc(2*ps);
#else
    emul = data;
#endif             
    emul =  (char*)((ps-1+(INT64)emul) & ~(ps-1));


    for( i=0; i < I_last; i++ )
    {
        CASSERT( InstructionTable[i].Code == i,("weird\n"));
        
        switch( InstructionTable[i].Type )
        {
        case IT_ULD:
        case IT_ILD:
            opcode = MAKE_OPC(InstructionTable[i].OpCode) +
                     MAKE_REGA(REG_V0) +
                     MAKE_REGB(REG_A0) +
                     MAKE_MDISP(0);
            break;
        case IT_FCM:
        case IT_FCP:
        case IT_FA4:
        case IT_FA8:
        case IT_FS4:
        case IT_FS8:
        case IT_FCV:
        case IT_FM4:
        case IT_FM8:
        case IT_FD4:
        case IT_FD8:
        case IT_FCS:
        case IT_FCN:
        case IT_FCE:
            opcode = MAKE_OPC(InstructionTable[i].OpCode) +
                     MAKE_IFUN(InstructionTable[i].FunCode) +    
                     MAKE_REGA(REG_F16) +
                     MAKE_REGB(REG_F17) +
                     MAKE_REGC(REG_F0);
            break;
        case IT_ICM:
        case IT_IOP:
        case IT_ICP:
        case IT_MLL:
        case IT_MLQ:
        case IT_MLH:
        case IT_EXT:
        case IT_LOG:
        case IT_SFT:
            opcode = MAKE_OPC(InstructionTable[i].OpCode) +
                     MAKE_IFUN(InstructionTable[i].FunCode) +    
                     MAKE_REGA(REG_A0) +
                     MAKE_REGB(REG_A1) +
                     MAKE_REGC(REG_V0);
            break;

        case IT_FLD:
            opcode = MAKE_OPC(InstructionTable[i].OpCode) +
                     MAKE_REGA(REG_F0) +
                     MAKE_REGB(REG_A0) +
                     MAKE_MDISP(0);
            break;
        default:
            opcode = ILLEGAL;
        }

        ((INT32 *)emul)[4*i] = opcode;
        ((INT32 *)emul)[4*i+1] = RET_OPCODE;
    }
    
    if( mprotect( emul, ps, PROT_READ | PROT_WRITE |  PROT_EXEC) )
    {
        FATAL(("a call to mprotect failed.\n"
               "Sometimes this can be remedied by relinking the program\n"
               "with a different order of object files\n"));
    }

    InfoStatus(stdout, "\t[EmulInitialize] Code area: %016lx\n",emul);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "EmulIOP"

extern INT64 EmulIOP(INDEX Code, INT64 RegValA, INT64 RegValB)  
{
    EVAL_IOP TheEvalIOP  =  (EVAL_IOP) & ((INT32 *)emul)[4*Code];
    
    return TheEvalIOP(RegValA,RegValB);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "EmulILD"

extern INT64 EmulILD(INDEX Code, INT64 RegVal)  
{
    EVAL_ILD TheEvalILD  =  (EVAL_ILD) & ((INT32 *)emul)[4*Code];
        
    return TheEvalILD(RegVal);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "EmulFLD"

extern double EmulFLD(INDEX Code, INT64 RegVal)  
{
    EVAL_FLD TheEvalFLD  =  (EVAL_FLD) & ((INT32 *)emul)[4*Code];
    return TheEvalFLD(RegVal);
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "EmulFOP"

extern double EmulFOP(INDEX Code, double RegValA, double RegValB)  
{
    EVAL_FOP TheEvalFOP  =  (EVAL_FOP) & ((INT32 *)emul)[4*Code];
    return TheEvalFOP(RegValA,RegValB);
}

/*=======================================================
=======================================================*/







