/*=======================================================
  Author:   R. Muth
  Version:  $Id: fun.c,v 1.23 1999/06/16 22:21:35 muth Exp $
  Comment:  This file contains the primitives for all function
            related computations
=======================================================*/

#include "global.h"
#include "flags.h"
#include "regs.h"
#include "functions.h"
#include "opcode.h"

/*=======================================================
@FUN
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunFindFreeRegs"

extern REGLIST AfunFindFreeRegs(INDEX ifun)
{
    INDEX exit;
    REGLIST avail;
    
    exit = AFUN_BBL_LAST(ifun);

    CASSERT( ABBL_TYPE(exit) == BT_EXIT,("weird\n"));

    avail = REGS_ALL_BUT_ZERO;
    avail &= ~AFUN_REGS_CHANGED(ifun);
    avail &= ~ABBL_OUT(exit);

    return avail;
}

/*=======================================================
@FUN
Assuming register reg is a saved register of ifun
find its stack offset. NEEDS WORK!
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunRegisterSaveOffset"

extern NUMBER AfunRegisterSaveOffset(INDEX ifun, REG reg)
{

    INDEX ibbl = AFUN_BBL_FIRST(ifun);
    INDEX iins;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( AINS_REG_C(iins) == reg )
            return INVALID;
        
        if( AINS_CODE(iins) != I_stq &&
            AINS_CODE(iins) != I_stt &&
            AINS_CODE(iins) != I_stq ) continue;
        if( AINS_REG_B(iins) != REG_SP ) continue;
        if( AINS_REG_A(iins) != reg ) continue;
        return AINS_DATA(iins);
    }
    return INVALID;
}

/*=======================================================
@FUN
Assuming register reg is a saved register of ifun
find its save instruction. NEEDS WORK!
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunRegSaveIns"

extern INDEX AfunRegSaveIns(INDEX ifun, REG reg)
{

    INDEX ibbl = AFUN_BBL_FIRST(ifun);
    INDEX iins;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( AINS_REG_C(iins) == reg )
            return INVALID;
        
        if( AINS_CODE(iins) != I_stq &&
            AINS_CODE(iins) != I_stt &&
            AINS_CODE(iins) != I_stq ) continue;
        if( AINS_REG_B(iins) != REG_SP ) continue;
        if( AINS_REG_A(iins) != reg ) continue;
        return iins;
    }
    return INVALID;
}

/*=======================================================
@FUN
Assuming register reg_saved is a saved register of ifun
avoid the save and restore operations by bypassing it
through reg_bypass
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunBypassSaveRegister"

extern void AfunBypassSaveRegister(INDEX ifun,REG reg_saved, REG reg_bypass)
{
    NUMBER offset = AfunRegisterSaveOffset(ifun,reg_saved);
    INDEX ibbl;
    INDEX iins;
    DATATYPE type;



    if( !VALID(offset) )
    {
        ERROR(("could not find stack offset for reg: %d\n",reg_saved));
        PrintBblSuper(stdout,AFUN_BBL_FIRST(ifun));
        return;
    }
    
    printf("fun %d [%s]  reg_saved %d  reg_bypass %d\n",
           ifun,AFUN_NAME(ifun),reg_saved,reg_bypass);

    /* do the store */
    
    ibbl = AFUN_BBL_FIRST(ifun);
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( AINS_DATA(iins) != offset ) continue;
        if( AINS_TYPE(iins) != IT_IST && AINS_TYPE(iins) != IT_FST ) continue;
        type = AinsDataType(iins);
        if( type != DT_TFLT && type != DT_GFLT && type != DT_QUAD ) continue;
        if( AINS_REG_B(iins) != REG_SP ) continue;
        if( AINS_REG_A(iins) != reg_saved ) continue;
        AinsMakeMove(iins,reg_saved,type,reg_bypass,type);
        break;
    }
    
    CASSERT( VALID(iins),("could not find reg %d  save in %d [%s]\n",
                          reg_saved,ifun,AFUN_NAME(ifun)));

    /* do the loads */
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_RET ) continue;
        if( ABBL_IS_CRIT_PATH(ibbl) ) continue;
        
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            if( AINS_REG_C(iins) != reg_saved )
            {
                if( AINS_DATA(iins) != offset ) break;
                if( AINS_TYPE(iins) != IT_ILD && AINS_TYPE(iins) != IT_FLD ) break;
                type = AinsDataType(iins);
                if( type != DT_TFLT && type != DT_GFLT && type != DT_QUAD ) break;
                if( AINS_REG_B(iins) != REG_SP ) break;
                AinsMakeMove(iins,reg_bypass,type,reg_saved,type);
                break;
            }
            
        }

        CASSERT( VALID(iins),("could not find reg %d load in %d [%s]\n",
                              reg_saved,ifun,AFUN_NAME(ifun)));
    }
}
/*=======================================================

=======================================================*/
#undef __PROC__
#define __PROC__ "AfunComputeChangedRegisters"

extern void AfunComputeChangedRegisters(INDEX ifun)
{
    /*
      This function computes the genuinely changed register ignoring
      loads that restore a register.
     */
    
    INDEX ibbl;
    INDEX iins;
    REGLIST changed;

    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch(ifun)
        {
            
        case AFUN_PSEUDO_PAL:
            AFUN_REGS_CHANGED(ifun) = REGS_PROC_INT_MAY_CHANGE;
            return;
        case AFUN_PSEUDO_CALL:
            AFUN_REGS_CHANGED(ifun) = REGS_PROC_INT_MAY_CHANGE | REGS_PROC_FLT_MAY_CHANGE;
            return;
	  case AFUN_PSEUDO_HELL_LITE:
            AFUN_REGS_CHANGED(ifun) = REGS_PROC_INT_MAY_CHANGE | REGS_PROC_FLT_MAY_CHANGE;
	    return;
        case AFUN_PSEUDO_HELL:
            AFUN_REGS_CHANGED(ifun) = REGS_ALL_BUT_ZERO;
            return;
        default:
            ERROR(("unexpected pseudo function: %d [%s]\n",ifun,AFUN_NAME(ifun)));
            AFUN_REGS_CHANGED(ifun) = REGS_ALL_BUT_ZERO;
            return;
        }
    }
    
    
    changed = REGS_NONE;
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            REG reg = AINS_REG_C(iins);
#if 0
            if( AINS_TYPE(iins) == IT_PAL )
            {
                changed = REGS_ALL_BUT_ZERO;
                break;
            }
#endif            
            if( REG2BIT(reg) & changed ) continue;
            changed |= REG2BIT(reg);
            if( ABBL_TYPE(ibbl) != BT_RET ) continue;
            if( AINS_CODE(iins) != I_ldq &&
                AINS_CODE(iins) != I_ldt &&
                AINS_CODE(iins) != I_ldg  ) continue;
            if( AINS_REG_B(iins) != REG_SP ) continue;
            if( AINS_DATA(iins) != AfunRegisterSaveOffset(ifun,reg) ) continue;
            changed &= ~REG2BIT(reg);                        
        }
    }

    AFUN_REGS_CHANGED(ifun) =  changed & ~(B_ZERO|B_FZERO);
}

/*=======================================================
@FUN
This function updates the AUNS_REGS_SAVED field
It computes the set of registers which are saved by ifun BUT NOT used.
Hence the return register should not not appear in this list!
A heuristic is used which should be somewhat conservative.
It would be nice to have more reliable information.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunComputeSavedRegisters"

extern void AfunComputeSavedRegisters(INDEX ifun)
{
    INDEX ibbl;
    INDEX iins;
    NUMBER count;
    NUMBER SavedRegOffset[64];
    REGLIST use;
    REGLIST def;
    REGLIST saved;
    REG reg;
    
    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch( ifun )
        {
        case AFUN_PSEUDO_PAL:
            AFUN_REGS_SAVED(ifun) = REGS_ALL_BUT_ZERO & (REGS_PROC_INT_PRESERVED|REGS_PROC_FLT_PRESERVED);
            return;
        case AFUN_PSEUDO_CALL:
            AFUN_REGS_SAVED(ifun) = REGS_ALL_BUT_ZERO & (REGS_PROC_INT_PRESERVED|REGS_PROC_FLT_PRESERVED);
            return;
        case AFUN_PSEUDO_HELL_LITE:
            AFUN_REGS_SAVED(ifun) = REGS_NONE;
            return;
        case AFUN_PSEUDO_HELL:
            AFUN_REGS_SAVED(ifun) = REGS_NONE;
            return;
	      
        default:
            ERROR(("unexpected pseudo function: %d [%s]\n",ifun,AFUN_NAME(ifun)));
            AFUN_REGS_SAVED(ifun) = REGS_NONE;
            return;
        }
    }

#if 0
    if( ifun > PRINT_OPT_NUM )
    {
        AFUN_REGS_SAVED(ifun) = REGS_NONE;
        return;
    }
#endif
    
    for(reg=0;reg<64;reg++)
    {
        SavedRegOffset[reg] = INVALID;
    }

    ibbl = AFUN_BBL_FIRST(ifun);
    saved = REGS_NONE;
    use  = REGS_NONE;
    def  = REGS_NONE;
    
    /* do the stores as a first approximation - we look in the very first bbl of the function*/

    
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( (AINS_CODE(iins) == I_stq || AINS_CODE(iins) == I_stt || AINS_CODE(iins) == I_stg) && AINS_REG_B(iins) == REG_SP )
        {
            reg = AINS_REG_A(iins);

            /* only if reg has not been used and not been defined
               we have a true save */ 

            if( (REG2BIT(reg) & use) == REGS_NONE &&
                (REG2BIT(reg) & def) == REGS_NONE)
            {
                if( saved & REG2BIT(reg) )
                {
                    saved &= ~REG2BIT(reg);
                    use |= REG2BIT(reg);  /* keep it from being looked at again */
                    SavedRegOffset[reg] = INVALID;
                    /* printf("unsaved reg %d in fun %d\n",reg,ifun); */
                }
                else
                {
                    saved |= REG2BIT(reg);
                    SavedRegOffset[reg] = AINS_DATA(iins);
                }
            }

            /* do not consider this a use ! */
        }
        else
        {
            saved &= ~AinsUse(iins) | def;
            for(reg=0;reg<64;reg++)
            {
                if( REG2BIT(reg) & saved ) continue;
                SavedRegOffset[reg] = INVALID;
            }
            def |= AinsDef(iins);
            use |= AinsUse(iins);
        }
        
    }
    
#if 0
    printf("first approximation %016lx for %d [%s]\n",saved,ifun,AFUN_NAME(ifun));
#endif

    /* now make sure that nothing used subsequently */

    saved &= ~ABBL_OUT(ibbl) | def;

    for(reg=0;reg<64;reg++)
    {
        if( REG2BIT(reg) & saved ) continue;
        SavedRegOffset[reg] = INVALID;
    }

#if 0
    printf("second approximation %016lx for %d [%s]\n",saved,ifun,AFUN_NAME(ifun));
#endif
    
    /*now check the loads - we look at all the return bbls*/

    count = 0;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_RET ) continue;
        if( ABBL_IS_CRIT_PATH(ibbl) ) continue;
        count++;

        def = REGS_NONE;
        saved = REGS_NONE;
        
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            reg =  AINS_REG_C(iins);
            if( REG2BIT(reg) & def ) continue;
            def |= AinsUse(iins) | AinsDef(iins);

            if( AINS_CODE(iins) != I_ldq &&
                AINS_CODE(iins) != I_ldt &&
                AINS_CODE(iins) != I_ldg  ) continue;
            if( AINS_REG_B(iins) != REG_SP ) continue;


            if( SavedRegOffset[reg] != AINS_DATA(iins) )
            {
                SavedRegOffset[reg] = INVALID;
                continue;
            }
            else
            {
                saved |= REG2BIT(reg);
            }
            
        }

        for(reg=0;reg<64;reg++)
        {
            if( REG2BIT(reg) & saved ) continue;
            SavedRegOffset[reg] = INVALID;
        }
    }

    if( count )
    {
        AFUN_REGS_SAVED(ifun) = REGS_NONE;

        for(reg=0;reg<64;reg++)
        {
            if( VALID(SavedRegOffset[reg]) )
                AFUN_REGS_SAVED(ifun) |= REG2BIT(reg);
        }

#if 0
        printf("saved %016lx %d [%s]\n",AFUN_REGS_SAVED(ifun),ifun,AFUN_NAME(ifun));
#endif
    }
    else
    {
        AFUN_REGS_SAVED(ifun) = REGS_NONE;
    }


}


/*=======================================================
@FUN
Assuming reg is a saved register in ifun delete
its save/restore pair.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunKillLoadStoreSaveRegister"


extern void AfunKillLoadStoreSaveRegister(INDEX ifun,REG reg)
{
    NUMBER offset = AfunRegisterSaveOffset(ifun,reg);
    BOOL flag;
    INDEX ibbl;
    INDEX iins;

    if( !VALID(offset) )
    {
        WARNING(("bad offset for reg: %d\n",reg));
        return;
    }
    

    ibbl = AFUN_BBL_FIRST(ifun);
    flag = 0;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( AINS_REG_C(iins) == reg ) break;
        
        if( AINS_DATA(iins) != offset ) continue;
        if( AINS_CODE(iins) != I_stq &&
            AINS_CODE(iins) != I_stt &&
            AINS_CODE(iins) != I_stg ) continue;
        if( AINS_REG_B(iins) != REG_SP ) continue;
        if( AINS_REG_A(iins) != reg ) continue;
        AinsKill(iins);
        flag = 1;
        break;
    }
    
    CWARNING( !flag,("could not find reg %d  save in %d [%s]\n",reg,ifun,AFUN_NAME(ifun)));


    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_RET ) continue;
        if( ABBL_IS_CRIT_PATH(ibbl) ) continue;

        flag = 0;
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            if( AINS_REG_C(iins) == reg )
            {
                if( AINS_DATA(iins) != offset ) break;
                if( AINS_CODE(iins) != I_ldq &&
                    AINS_CODE(iins) != I_ldt &&
                    AINS_CODE(iins) != I_ldg  ) continue;
                if( AINS_REG_B(iins) != REG_SP ) break;
                AinsKill(iins);
                flag = 1;
                break;
            }
            
        }

        CWARNING( !flag,("could not find reg %d load in %d [%s]\n",reg,ifun,AFUN_NAME(ifun)));
    }
}

/*=======================================================
@FUN
Find the registers in ifun which are neither used nore defined
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunScratchRegisters"

extern REGLIST AfunScratchRegisters(INDEX ifun)
{
    INDEX ibbl;
    INDEX iins;
    REGLIST scratch = ~ABBL_OUT(AFUN_BBL_LAST(ifun));

    scratch &= ~B_ZERO;
    scratch &= ~B_FZERO;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) == BT_CALL )
            return REGS_NONE;
        
        FOREACH_BBL_INS(ibbl,iins)
        {
            scratch &= ~AinsUse(iins);
            scratch &= ~AinsDef(iins);
        }
    }
    

    return scratch;
}

/*=======================================================
@FUN
Return True if ifun does not call any other function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunIsLeaf"

extern BOOL AfunIsLeaf(INDEX ifun)
{
    INDEX ibbl;

    if( AfunHasEsc2(ifun) )
        return 0;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) == BT_CALL ) return 0;
        if( ABBL_TYPE(ibbl) == BT_PAL ) return 0;      
    }

    return 1;
}
/*=======================================================
@FUN
Return True if ifun can be loop-optimized without
making the optimizer go into an infinite loop
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunIsLoopAnalyzable"

extern BOOL AfunIsLoopAnalyzable(INDEX ifun)
{
    if( AFUN_IS_PSEUDO(ifun) ) return 0;
    if( AFUN_IS_SETJMP(ifun) ) return 0;
    if( AFUN_IS_LONGJMP(ifun) ) return 0;
    
    if( AfunHasEsc2(ifun) ) return 0;
    if( AfunHas2EscNotFirst(ifun) ) return 0;
    if( AfunNBbls(ifun) != AfunReachableFirst(ifun) ) return 0;
    if( AfunNBbls(ifun) != AfunReachableLast(ifun) ) return 0;
    
    return 1;
}
/*=======================================================
@FUN
Return True if all  bbls of ifun can be reached from the first (init) bbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunReachableFirst"

extern NUMBER AfunReachableFirst(INDEX ifun)
{
    INDEX ibbl;
    INDEX iedg;
    BOOL flag;
    NUMBER count;
    
    FOREACH_FUN_BBL(ifun,ibbl)
        ABBL_FLAGS(ibbl) &= ~F_MARKED;

    ABBL_FLAGS(AFUN_BBL_FIRST(ifun)) |= F_MARKED;
    count = 1;
    
    do
    {
        flag = 0;
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( !ABBL_IS_MARKED(ibbl) ) continue;
            if( ABBL_IS_LAST(ibbl) ) continue;
            
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) ) continue;
                if( ABBL_IS_MARKED(AEDG_TAIL(iedg)) ) continue;
                flag = 1;
                count++;
                ABBL_FLAGS(AEDG_TAIL(iedg)) |= F_MARKED;
            }
        }
        
    }
    while( flag );
    

    return count;
}
/*=======================================================
@FUN
Return True if all  bbls of ifun can be (back) reached from the last (exit) bbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunReachableLast"

extern NUMBER AfunReachableLast(INDEX ifun)
{
    INDEX ibbl;
    INDEX iedg;
    BOOL flag;
    NUMBER count;
    
    FOREACH_FUN_BBL(ifun,ibbl)
        ABBL_FLAGS(ibbl) &= ~F_MARKED;

    ABBL_FLAGS(AFUN_BBL_LAST(ifun)) |= F_MARKED;
    count = 1;
    
    do
    {
        flag = 0;
        FOREACH_FUN_BBL_R(ifun,ibbl)
        {
            if( !ABBL_IS_MARKED(ibbl) ) continue;
            if( ABBL_IS_FIRST(ibbl)  ) continue;
               
            FOREACH_BBL_PRED(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) ) continue;
                if( ABBL_IS_MARKED(AEDG_HEAD(iedg)) ) continue;
                flag = 1;
                count++;
                ABBL_FLAGS(AEDG_HEAD(iedg)) |= F_MARKED;
            }
        }
    }
    while( flag );
    
    return count;
}
/*=======================================================
@FUN
Return the number of bbls of type BT_SWITCH in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNSwitches"

extern NUMBER AfunNSwitches(INDEX ifun)
{
    INDEX ibbl;
    NUMBER count=0;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) == BT_SWITCH )
            count++;
    }

    return count;
}
/*=======================================================
@FUN
Return the number of critical bbls of type BT_SWITCH in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNSwitchesCritical"

extern NUMBER AfunNSwitchesCritical(INDEX ifun)
{
    INDEX ibbl;
    NUMBER count=0;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        if( ABBL_TYPE(ibbl) == BT_SWITCH )
            count++;
    }

    return count;
}
/*=======================================================
@FUN
Return the number of bbls of type BT_JUMP in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNJumps"

extern NUMBER AfunNJumps(INDEX ifun)
{
    INDEX ibbl;
    NUMBER count=0;
        
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) == BT_JUMP )
            count++;
    }

    return count;
}
/*=======================================================
@FUN
Return the number of critical bbls of type BT_JUMP in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNJumpsCritical"

extern NUMBER AfunNJumpsCritical(INDEX ifun)
{
    INDEX ibbl;
    NUMBER count=0;
        
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        if( ABBL_TYPE(ibbl) == BT_JUMP )
            count++;
    }
    return count;
}
/*=======================================================
@FUN
Return True if ifun has outgoing escaping edges
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunHasEsc2"

extern BOOL AfunHasEsc2(INDEX ifun)
{
    INDEX ibbl;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX iedg;

        if( ABBL_IS_LAST(ibbl) ) continue;
        
        FOREACH_BBL_SUCC(ibbl,iedg)
        {
            
            if( !AEDG_IS_INTER(iedg) ) continue;
            if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
            if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;
            return 1; 
        }
    }

    return 0;
}
/*=======================================================
@FUN
Return True if ifun has outgoing escaping edges from critical bbls
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunHasEsc2Critical"

extern BOOL AfunHasEsc2Critical(INDEX ifun)
{
    INDEX ibbl;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX iedg;

        if( ABBL_IS_LAST(ibbl) ) continue;
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        
        FOREACH_BBL_SUCC(ibbl,iedg)
        {
            
            if( !AEDG_IS_INTER(iedg) ) continue;
            if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
            if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;
            return 1; 
        }
    }

    return 0;
}

/*=======================================================
@FUN
Return True if ifun has incoming escaping edges
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunHas2Esc"

extern BOOL AfunHas2Esc(INDEX ifun)
{
    INDEX ibbl;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX iedg;
        if( ABBL_IS_LAST(ibbl) ) continue;
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            if( !AEDG_IS_INTER(iedg) ) continue;
            if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
            /*
              NEEDS WORK: THIS IS A QUICK HACK TO COPE WITH
               PATH INLINING. USUALLY THERE ARE NO CALLS TO NON
               FIRST BBLS.
               */
            if( AEDG_TYPE(iedg) == ET_FUN_CALL && ABBL_IS_FIRST(ibbl) ) continue;

            return 1;
        }
    }

    return 0;
}
/*=======================================================
@FUN
Return True if ifun has incoming escaping edges to bbls other than the init bbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunHas2EscNotFirst"

extern BOOL AfunHas2EscNotFirst(INDEX ifun)
{
    INDEX ibbl;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX iedg;
        if( ABBL_IS_FIRST(ibbl) ) continue;
        if( ABBL_IS_LAST(ibbl) ) continue;
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            if( !AEDG_IS_INTER(iedg) ) continue;
            if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
            /*
              NEEDS WORK: THIS IS A QUICK HACK TO COPE WITH
               PATH INLINING. USUALLY THERE ARE NO CALLS TO NON
               FIRST BBLS.
             */
            if( AEDG_TYPE(iedg) == ET_FUN_CALL && ABBL_IS_FIRST(ibbl) ) continue;

            return 1;
        }
    }

    return 0;
}
/*=======================================================
@FUN
obsolete
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunIsSetjmp"

extern BOOL AfunIsSetjmp(INDEX ifun)
{
    INDEX exit = AFUN_BBL_LAST(ifun);
    INDEX iedg;

    FOREACH_BBL_PRED(exit,iedg)
        if( AEDG_TYPE(iedg) == ET_SETLONGJMP ) return 1;
    
    return 0;
}
/*=======================================================
@FUN
obsolete
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunIsLongjmp"

extern BOOL AfunIsLongjmp(INDEX ifun)
{
    INDEX exit = AFUN_BBL_LAST(ifun);
    INDEX iedg;

    FOREACH_BBL_SUCC(exit,iedg)
        if( AEDG_TYPE(iedg) == ET_SETLONGJMP ) return 1;
    
    return 0;
}
/*=======================================================
@FUN
Return the return register of ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunReturnRegister"     

extern REG AfunReturnRegister(INDEX ifun)
{
    INDEX ibbl;
    INDEX iins;
    REG   ret = INVALID;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        iins = ABBL_INS_LAST(ibbl);
        if( !VALID(iins) ) continue;
        if( AINS_CODE(iins) != I_ret ) continue;
        if( ret == INVALID )
            ret = AINS_REG_B(iins);
        CWARNING( ret != AINS_REG_B(iins),
                 ("different return registers in function %d [%s]\n",ifun,AFUN_NAME(ifun)));
    }

    return ret;
}

/*=======================================================
@FUN
Return the number of call edges calling ifun from critical bbls,
does include incoming escapes.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNCriticalCallSites"

extern NUMBER AfunNCriticalCallSites(INDEX ifun)
{
    INDEX iedg;
    NUMBER count = 0;
    INDEX init = AFUN_BBL_FIRST(ifun);

    if( !VALID(init) ) return count;
    
    FOREACH_BBL_PRED(init,iedg)
    {
        if( !AEDG_IS_INTER(iedg) ) continue;
        if( ABBL_IS_CRITICAL(AEDG_HEAD(iedg)) ) count++;
    }

    return count;
}

/*=======================================================
@FUN
Return the number of call edges calling ifun from  hot bbls,
does include incoming escapes.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNHotCallSites"

extern NUMBER AfunNHotCallSites(INDEX ifun)
{
    INDEX iedg;
    NUMBER count = 0;
    INDEX init = AFUN_BBL_FIRST(ifun);

    if( !VALID(init) ) return count;
    
    FOREACH_BBL_PRED(init,iedg)
    {
        if( !AEDG_IS_INTER(iedg) ) continue;
        if( ABBL_IS_HOT(AEDG_HEAD(iedg)) ) count++;
    }

    return count;
}

/*=======================================================
@FUN
Return the number of call edges calling ifun,
does include incoming escapes.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNCallSites"

extern NUMBER AfunNCallSites(INDEX ifun)
{
    NUMBER count=0;
    INDEX iedg;
    INDEX  ibbl = AFUN_BBL_FIRST(ifun);
    CASSERT( VALID(ifun),("Bad argument fun: %d\n",ifun));
    CASSERT( VALID(ibbl),("Bad first bbl in fun %d\n",ifun));

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        if( !AEDG_IS_INTER(iedg) ) continue;
        count++;
    }
    
    return count;
}
/*=======================================================
@FUN
Return the number of critical inss in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNInssCritical"

extern NUMBER AfunNInssCritical(INDEX ifun)
{
    INDEX  ibbl;
    NUMBER i=0;
    CASSERT( VALID(ifun),("Bad argument fun: %d\n",ifun));
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_FLAGS(ibbl) & FB_CRITICAL )
            i+=AbblNInss(ibbl);
    }
    
    return i;
}
/*=======================================================
@FUN
Return the number of inss in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNInss"

extern NUMBER AfunNInss(INDEX ifun)
{
    INDEX  ibbl;
    NUMBER i=0;
    CASSERT( VALID(ifun),("Bad argument fun: %d\n",ifun));
    FOREACH_FUN_BBL(ifun,ibbl) i+=AbblNInss(ibbl);
    return i;
}
/*=======================================================
@FUN
Return the number of critical bbls in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNBblsCritical"

extern NUMBER AfunNBblsCritical(INDEX ifun)
{
    INDEX  ibbl;
    NUMBER i=0;
    CASSERT( VALID(ifun),("Bad argument fun: %d\n",ifun));
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_IS_CRITICAL(ibbl) )
            i++;
    }
    
    return i;
}
/*=======================================================
@FUN
Return the number of hot bbls in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNBblsHot"

extern NUMBER AfunNBblsHot(INDEX ifun)
{
    INDEX  ibbl;
    NUMBER i=0;
    CASSERT( VALID(ifun),("Bad argument fun: %d\n",ifun));
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_IS_HOT(ibbl) )
            i++;
    }
    
    return i;
}
/*=======================================================
@FUN
Return the number of bbls in ifun
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNBbls"

extern NUMBER AfunNBbls(INDEX ifun)
{
    INDEX  ibbl;
    NUMBER i=0;
    CASSERT( VALID(ifun),("Bad argument fun: %d\n",ifun));
    FOREACH_FUN_BBL(ifun,ibbl) i++;
    return i;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunInit"

static void AfunInit(INDEX ifun)
{
    CASSERT( VALID(ifun), ("Bad argument fun: %d\n",ifun));
    
    AFUN_FLAGS(ifun) = 0;
    AFUN_NAME(ifun) = "no_name_yet";
    AFUN_PREV(ifun) = INVALID;
    AFUN_NEXT(ifun) = INVALID;
    AFUN_BBL_FIRST(ifun) = INVALID;
    AFUN_BBL_LAST(ifun) = INVALID;
    AFUN_STK(ifun) = INVALID;
    AFUN_GP(ifun) = INVALID;      
    AFUN_RET_REG(ifun) = INVALID; 
    AFUN_REGS_CHANGED(ifun) = REGS_ALL_BUT_ZERO;
    AFUN_REGS_SAVED(ifun) = REGS_NONE;

}

/*=======================================================
@FUN
Kill the function including it bbls edgs and ins.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunKill"

extern void AfunKill(INDEX ifun)
{
    INDEX ibbl;
    CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun),("Bad argument fun: %d\n",ifun));

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_FUN(ibbl) != ifun )
        {
            WARNING(("Attempt to delete bbl %d not belonging to fun %d\n",ibbl,ifun));
            continue;
        }
        else
            AbblKill(ibbl);
    }
    
    AFUN_FLAGS(ifun) |= F_KILLED;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNew"

static INDEX AfunNew()
{
    CFATAL(CONFIG_UFUN >= CONFIG_NFUN,("not enough functions\n"));
    AfunInit(CONFIG_UFUN);
    return(CONFIG_UFUN++);
}

/*=======================================================
@FUN
Create a new essentially empty function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNewFancy"

extern INDEX AfunNewFancy()
{
    INDEX ibbl;
    INDEX ifun;

    ifun = AfunNew();
    ibbl = AbblNew();
    ABBL_FUN(ibbl) = ifun;
    ABBL_TYPE(ibbl) = BT_EXIT;
    AFUN_BBL_FIRST(ifun) = ibbl;
    AFUN_BBL_LAST(ifun) = ibbl;
    AFUN_MASTER(ifun) = INVALID;
    return ifun;
}
/*=======================================================
=======================================================*/
