/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.constants.c,v 1.13 1999/06/15 21:48:00 muth Exp $
  Comment:  This file contains the optimization code for constants
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

/*=======================================================
@OPT
Compute constants faster.
use lda, ldah with rzero or gp
substitute a load if instruction is expensive
TODO: use lda/ldah pairs for 32 bit instructions
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstGenOne"

extern BOOL OptConstGenOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constgenone");
    BOOL change = 0;
    INDEX ibbl;
    INDEX iins;
    INDEX jins;
    NUMBER ccount=0;
    NUMBER ccounth=0;
    REG reg;
    INT64 value;
    INT64 value2;
    INT32 hi,lo;

    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( ibbl <= 6 ) continue;   /* GROSS HACK NEEDS WORK*/

        if( !(AINS_FLAGS(iins) & FI_EVAL_C) ) continue;
        reg = AINS_REG_C(iins);
        if( reg == REG_ZERO || reg == REG_FZERO  || reg == REG_GP ) continue;
        value = AINS_C_EVAL(iins);

     
        /* check whether it is in the text segement */

        if( AddressIsCode(value) ) continue;
        if( AinsHasSideEffect(iins) ) continue;
        if( REG_IS_INT(reg) )
        {
            /* ====================================================================== */
            
            if( AINS_CODE(iins) == I_lda && AINS_REG_B(iins) == REG_ZERO ) continue;    


            if( -(1L<<15) <=  value  && value < (1L<<15) )
            {
                AinsMakeLda(iins,REG_ZERO,reg,value);
                goto did_something;
            }

            /* ====================================================================== */
            
            if( AINS_CODE(iins) == I_ldah && AINS_REG_B(iins) == REG_ZERO ) continue;

            if( (value & 0xffff) == 0 && ((-(1L<<15))<<16) <= value && value < (1L<<31) )
            {
                AinsMakeLdah(iins,REG_ZERO,reg,value>>16);
                goto did_something;
            }

            /* ====================================================================== */
            value2 = value - SysGetGlobalPointer();
            /* ====================================================================== */
            
            if( AINS_CODE(iins) == I_lda )
            
            if( -(1L<<15) <=  value2  && value2 < (1L<<15) )
            {
                AinsMakeLda(iins,REG_GP,reg,value2);
                goto did_something;
            }


            /* ====================================================================== */
            
            if( AINS_TYPE(iins) == IT_LDA || AINS_TYPE(iins) == IT_FCS ) continue;

            if( -(1L<<30) <=  value  && value < (1L<<30) )
            {
                hi = value >> 16;
                lo = value & 0xffff;
                if( lo & 0x8000 )
                {
                    lo |= 0xffff0000;
                    hi += 1;
                }

                CASSERT( (hi << 16) + lo == value,("weird\n"));
                jins = AinsNew();
#if 0
                HURRAH(("did 32 constant: %16lx with instruction %d\n",value,iins));
                PrintBblSuper(stdout,AINS_BBL(iins));
#endif
                AinsMakeLdah(jins,REG_ZERO,reg,hi);
                AINS_BBL(jins) = AINS_BBL(iins);
                AinsInsertBefore(jins,iins);                
                AinsMakeLda(iins,reg,reg,lo);
                goto did_something;
            }
            
            /* ====================================================================== */
            if( -(1L<<30) <=  value2  && value2 < (1L<<30) )
            {
                hi = value2 >> 16;
                lo = value2 & 0xffff;
                if( lo & 0x8000 )
                {
                    lo |= 0xffff0000;
                    hi += 1;
                }

                CASSERT( (hi << 16) + lo == value2,("weird\n"));
                jins = AinsNew();
#if 0
                HURRAH(("did 32 constant: %16lx with instruction %d\n",value,iins));
                PrintBblSuper(stdout,AINS_BBL(iins));
#endif
                AinsMakeLdah(jins,REG_GP,reg,hi);
                AINS_BBL(jins) = AINS_BBL(iins);
                AinsInsertBefore(jins,iins);                
                AinsMakeLda(iins,reg,reg,lo);
                goto did_something;
            }

            /* ====================================================================== */
            
            if( AINS_TYPE(iins) == IT_MLL || AINS_TYPE(iins) == IT_MLQ || AINS_TYPE(iins) == IT_MLH )
            {
                POTENTIAL(("constant ins %d [%s] val: %016lx bbl: %d (%ld)\n",iins,AINS_NAME(iins),value,ibbl,ABBL_COUNT(ibbl)));
#if 0
                PrintBblSuper(stdout,ibbl);
#endif
            }
                
        }
        else
        {
            double d;

            if( value == 0 || value == (1L<<63) )
            {
                if( AINS_CODE(iins) == I_cpys || AINS_CODE(iins) == I_cpysn ) continue;
            }
            
            if( AINS_TYPE(iins) == IT_FLD ) continue;
            if( AINS_TYPE(iins) == IT_FCS ) continue;
            if( ABBL_IS_HOT(AINS_BBL(iins)) )
            {
                *((INT64 *)&d) = value;
                POTENTIAL(("constant ins %d [%s] val: %8f  bbl: %d (%ld)\n",iins,AINS_NAME(iins),d,ibbl,ABBL_COUNT(ibbl)));
#if 0
                PrintBblSuper(stdout,ibbl);
#endif
            }
        }


        continue;
        
did_something:
        AINS_FLAGS(iins) |= FI_EVAL_C;
        AINS_C_EVAL(iins) = value;
        ccount++;    
        if( ABBL_IS_HOT(ibbl) )
            ccounth++;
        
        if( verbose )
        {
            fprintf(fp,"\t[*] ins %d created %s for value: %16lx\n",iins,AINS_NAME(iins),value);
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (ccount > 0);

    InfoStatus(fp,"\t[OptConstgenone%c] Optimized Constants: %d#%d {%d ms}\n",CHANGE(change),ccount,ccounth,time);

    return change;
}


/*=======================================================
@OPTIMIZATION
Sophisticated constant propagation.
We resuse a constant in some other register to (hopefully) reduce 
register usage.
This is only done on the basic block level.
We do not kill instructions but expect useless code to be eliminated
by other optimizations.
We also allow construction of values from other using
either explict lda's or doing it implicitly for
memory format instruction. Care must be taken to not construct
values that point into code sections since those values
may change!!!
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstGenTwo"

extern BOOL OptConstGenTwo(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constgentwo"); 
    BOOL   change = 0;
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    INDEX  ibbl;
    REG    reg;
    
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        INDEX  iins;
        
        REGLIST hasValue;
        INT64   CurrValue[64];
        NUMBER  LastDef[64];
        NUMBER  no;

        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;
         
        no = 0;

        for(reg=0;reg<64;reg++)
            LastDef[reg]=-666;
        
        CurrValue[REG_ZERO] = 0;
        CurrValue[REG_FZERO] = 0;
        CurrValue[REG_GP] = SysGetGlobalPointer();

        hasValue = B_ZERO | B_FZERO | B_GP;
        
        
        FOREACH_BBL_INS(ibbl,iins)
        {
            REG ra = AINS_REG_A(iins);
            REG rb = AINS_REG_B(iins);
            REG rc = AINS_REG_C(iins);
            INT64 va = AINS_A_EVAL(iins);
            INT64 vb = AINS_B_EVAL(iins);
            INT64 vc = AINS_C_EVAL(iins);
            NUMBER oldest;
            REG    newreg;
            BOOL   isMemFormat;
            
            no++;
            
            /* ====================================================================== */
            /* REGISTER A */
            /* ====================================================================== */
            
            if( (AINS_FLAGS(iins) & FI_EVAL_A) &&
                ra != REG_ZERO && ra != REG_LIT && ra != REG_FZERO && ra != REG_GP && ra != REG_SP &&
                AINS_TYPE(iins) != IT_CST )
            {
                if( hasValue & REG2BIT(ra) )
                {
                    CASSERT(CurrValue[ra]==va,("ra %d new %lx old %lx\n",ra,CurrValue[ra],va));
                }
                else
                {
                    hasValue |= REG2BIT(ra);
                    CurrValue[ra] = va;
                }

                                
                oldest = no;
                newreg = ra;
                
                REGSET_FOREACH_REG(hasValue & (REG_IS_INT(ra) ? REGS_INT : REGS_FLT) ,reg) 
                {
                    if( CurrValue[reg] != va ) continue;
                    if( oldest > LastDef[reg] )
                    {
                        oldest = LastDef[reg];
                        newreg = reg;
                    }
                }
                    
                if( newreg != ra )
                {
                    if( verbose )
                    {
                        fprintf(fp, "\t[*] bbl: %d ins: %d rename reg a %d -> %d (%016lx)\n",ibbl,iins,ra,newreg,va);
                        PrintIns(stdout,iins);
                        /*
                          PrintBblSuper(stdout,AINS_BBL(iins));
                        */
                    }
                    
                    AINS_REG_A(iins) = newreg;
                    if( ABBL_IS_HOT(ibbl) )
                        pcounth++;
                    else
                        pcount++;
                }
            }

            /* ====================================================================== */
            /* REGISTER B (more oppportunities than with reg a because we can do an addition with
               a constant in the instruction) */
            /* ====================================================================== */
            
            if( (AINS_FLAGS(iins) & FI_EVAL_B) &&
                rb != REG_ZERO && rb != REG_LIT && rb != REG_FZERO && rb != REG_GP && rb != REG_SP )
            {
                if( hasValue & REG2BIT(rb) )
                {
                    CASSERT(CurrValue[rb]==vb,("rb %d new %lx old %lx\n",rb,CurrValue[rb],vb));
                }
                else
                {
                    hasValue |= REG2BIT(rb);
                    CurrValue[rb] = vb;
                }
                    
                oldest = no;
                newreg = rb;

                /*
                  we do not mess a lot with lda/ldah since if the base
                  register know the target register will be known as well
                  */
                
                if( (AINS_IS_LOAD(iins) || AINS_IS_STORE(iins)) && !AddressIsCode(vb) )
                    isMemFormat = 1;
                else
                    isMemFormat = 0;

                REGSET_FOREACH_REG(hasValue & (REG_IS_INT(rb) ? REGS_INT : REGS_FLT),reg) 
                {
                    if( isMemFormat )
                    {
                        if( vb + AINS_DATA(iins) - CurrValue[reg] >= (1<<15) ) continue;
                        if( vb + AINS_DATA(iins) - CurrValue[reg] <  -(1<<15) ) continue;
                    }
                    else
                    {
                        if( vb != CurrValue[reg] ) continue;
                    }
                    
                    if( oldest > LastDef[reg] )
                    {
                        oldest = LastDef[reg];
                        newreg = reg;
                    }
                }
                
                if( newreg != rb )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] bbl: %d ins: %d rename reg b %d (%016lx) -> %d (%016lx)\n",
                                ibbl,iins,rb,vb,newreg,CurrValue[newreg]);
                        PrintIns(stdout,iins);
                        /*
                        PrintBblSuper(stdout,AINS_BBL(iins));
                        */
                    }

                    
                    if( isMemFormat )
                    {
                        AINS_DATA(iins) =  vb + AINS_DATA(iins) - CurrValue[newreg];
                        AINS_B_EVAL(iins) = CurrValue[newreg];
                    }
                    
                    AINS_REG_B(iins) = newreg;
                    
                    if( ABBL_IS_HOT(ibbl) )
                        pcounth++;
                    else
                        pcount++;
                }
                
            }
            

            /* ====================================================================== */
            /* REGISTER C */
            /* ====================================================================== */
            
            if( (AINS_FLAGS(iins) & FI_EVAL_C) )
            {
                if( rc != REG_ZERO && rc != REG_LIT && rc != REG_FZERO && rc != REG_GP && rc != REG_SP )                
                {
#if 1
                    if( !AddressIsCode(vc) && REG_IS_INT(rc) )
                        isMemFormat = 1;
                    else
                        isMemFormat = 0;
                    
                    oldest = no;
                    newreg = INVALID;

                    REGSET_FOREACH_REG(hasValue & (REG_IS_INT(rc) ? REGS_INT: REGS_FLT),reg) 
                    {
                        if( isMemFormat )
                        {
                            INT64 difference =  vc - CurrValue[reg];
                            if( difference >= (1<<15) ) continue;
                            if( difference <  -(1<<15) ) continue;
                        }
                        else
                        {
                            if( vc != CurrValue[reg] ) continue;
                        }
                        
                        if( oldest > LastDef[reg] )
                        {
                            oldest = LastDef[reg];
                            newreg = reg;
                        }
                    }
                        
                    if( VALID(newreg) &&
                        AINS_TYPE(iins) != IT_FCS &&
                        /* AINS_TYPE(iins) != IT_FLD &&   float loads are faster than fcpys */
                        (AINS_TYPE(iins) != IT_LDA || (rb!=newreg && rb!=REG_GP && rb!=REG_ZERO) ))
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] bbl: %d  ins: %d rename reg c %d (%016lx) -> %d (%016lx)\n",
                                    ibbl,iins,rc,vc,newreg,CurrValue[newreg]);
                            PrintIns(stdout,iins);
                            /*
                              PrintBblSuper(stdout,AINS_BBL(iins));
                              */
                        }
                        
                        if( isMemFormat )
                        {
                            AinsMakeLda(iins,newreg,rc, vc - CurrValue[newreg]);
                        }
                        else
                        {
                            if( REG_IS_INT(rc) )
                            {
                                AinsMakeLda(iins,newreg,rc, 0);
                            }
                            else
                            {
                                AinsMakeCpys(iins,newreg,newreg,rc);
                                
                            }
                        }
                        
                        AINS_FLAGS(iins) |= FI_EVAL_C;
                        AINS_C_EVAL(iins) = vc;
                        
                        
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;
                        else
                            pcount++;
                    }
                    
                    hasValue |= REG2BIT(rc);
                    CurrValue[rc] = vc;
                    LastDef[rc] = no;
                }                    
#endif
            }
            else
            {
                hasValue &= ~REG2BIT(rc);
            }
        }
    }

               
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = pcount>0;

    InfoStatus(fp,"\t[OptConstgentwo%c] Modified ins: %d#%d {%d ms}\n",CHANGE(change),pcount,pcounth,time);

    return change;
}

/*=======================================================
=======================================================*/



/*=======================================================
@FUN

one and two set should be disjoint!

@END
=======================================================*/


static NUMBER OpTransNegateOpOne[] =
{
    I_blbc,I_blbs,
    I_blbs,I_blbc,
    I_beq,I_bne,
    I_bne,I_beq,

    /*  these should not occur */
#if 0
    I_ble,I_bgt, 
    I_bgt,I_ble,
    I_bge,I_blt,
    I_blt,I_bge,
#endif
    
    I_cmovne,I_cmoveq,
    I_cmoveq,I_cmovne,
    I_cmovgt,I_cmovle,
    I_cmovle,I_cmovgt,
    I_cmovlt,I_cmovge,
    I_cmovge,I_cmovlt,
    I_cmovlbc,I_cmovlbs,
    I_cmovlbs,I_cmovlbc,

    /* these require operand switching */
#if 0    
    I_bis,I_ornot,
    I_xor,I_eqv,
    I_and.I_bic,
#endif
    INVALID,INVALID
};


static NUMBER OpTransNegateOpTwo[] =
{
#if 0
    I_bis,I_ornot,
    I_ornot,I_bis,
    I_xor,I_eqv,
    I_eqv,I_xor,
    I_and,I_bic,
    I_bic,I_and,
#endif
    INVALID,INVALID
};


static NUMBER OpTransInvertOpOne[] =
{
    I_beq,I_beq,   /* no change !!! */
    I_bne,I_bne,   /* no change !!! */
    I_ble,I_bgt,
    I_bgt,I_ble,
    I_bge,I_blt,
    I_blt,I_bge,

    I_cmovne,I_cmovne, /* no change !!! */
    I_cmoveq,I_cmoveq, /* no change !!! */
    I_cmovge,I_cmovle,
    I_cmovle,I_cmovge,
    I_cmovlt,I_cmovgt,
    I_cmovgt,I_cmovlt,

    /* these require operand switching */
#if 0    
    I_addq,I_subq,
    I_addl,I_subl,
#endif
    INVALID,INVALID
};


static NUMBER OpTransInvertOpTwo[] =
{
    I_addq,I_subq,
    I_subq,I_addq,
    I_addl,I_subl,
    I_subl,I_addl,

    INVALID,INVALID
};



static NUMBER OpcodeTranslate(NUMBER opcode,NUMBER *OpTrans)
{
    while( *OpTrans != INVALID )
    {
        if( *OpTrans == opcode )
        {
            return OpTrans[1];
        }

        OpTrans+=2;
    }
    
    return INVALID;
}



/*=======================================================
@FUN
@END
=======================================================*/
static BOOL AinsTransform(INDEX iins_the,NUMBER *OpOneTrans, NUMBER *OpTwoTrans, BOOL DoTrans )
{
    INDEX reg;
    INDEX jins;
    NUMBER newop;
    
    reg = AINS_REG_C(iins_the);

    
    for(jins=AINS_NEXT(iins_the); VALID(jins); jins = AINS_NEXT(jins))
    {
        if( AINS_REG_B(jins) == reg )
        {
            newop = OpcodeTranslate(AINS_CODE(jins),OpTwoTrans);
            if( !VALID(newop) )
            {
                CERROR( DoTrans,("could not do transform!\n"));
                return 0;
            }

            if( DoTrans )
            {
                AINS_CODE(jins) = newop;
            }
            
        }
        
        if( AINS_REG_A(jins) == reg )
        {
            newop = OpcodeTranslate(AINS_CODE(jins),OpOneTrans);
            if( !VALID(newop) )
            {
                CERROR( DoTrans,("could not do transform!\n"));
                return 0;
            }
            
            if( DoTrans )
            {
                AINS_CODE(jins) = newop;
            }
        }
        if
            ( AINS_REG_C(jins) == reg )
        {
            if( AINS_TYPE(jins) == IT_ICM || AINS_TYPE(jins) == IT_FCM ) return 0;
            else break;  /* ok we are fine */
        }
    }

    if( !VALID(jins) )
    {
        if( ABBL_OUT(AINS_BBL(iins_the)) & REG2BIT(reg) )
        {
            CERROR( DoTrans,("could not do transform!\n"));
            return 0;
        }
        
    }

    return 1;
}

/*=======================================================
@FUN
@END
=======================================================*/

static void AinsSwitchOperands(INDEX iins)
{
    REG reg_tmp;
    INT64 val_tmp;
    
    CASSERT( AINS_REG_B(iins) != REG_LIT,("cannot switch iins %d\n",iins));

    reg_tmp = AINS_REG_A(iins);
    AINS_REG_A(iins) = AINS_REG_B(iins);
    AINS_REG_B(iins) = reg_tmp;

    val_tmp = AINS_A_EVAL(iins);
    AINS_A_EVAL(iins) = AINS_B_EVAL(iins);
    AINS_B_EVAL(iins) = val_tmp;


    if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            AINS_FLAGS(iins) &= ~FI_EVAL_A;
            AINS_FLAGS(iins) |= FI_EVAL_B;            
        }
    }
    else
    {
        if( AINS_FLAGS(iins) & FI_EVAL_B )
        {
            AINS_FLAGS(iins) &= ~FI_EVAL_B;
            AINS_FLAGS(iins) |= FI_EVAL_A;            
        }
    }
}


/*=======================================================
@OPT
Some alpha instruction can take a small immediate value instead of register B.
Now suppose that register A evaluates to such a small value, this optimization
tries to switch register A and B so that we can use the immediate value feature.
Example: xor r1,r2,r3 (we knwo that r1 contains the value 66)
         ----> xor r2,#66,r3 (using commutativity)
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstUseOne"

extern NUMBER OptConstUseOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constuseone");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;

    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    
    FOREACH_LIVE_FUN(ifun)
    FOREACH_FUN_BBL(ifun,ibbl)
    FOREACH_BBL_INS(ibbl,iins)
    {
        
        /* use commutativity to make the second operand the constant one, do not switch REG_ZERO */
        
        if( (AINS_FLAGS(iins) & FI_EVAL_A)  &&
            !(AINS_FLAGS(iins) & FI_EVAL_B) &&
            AINS_REG_A(iins) != REG_ZERO &&
            AINS_REG_B(iins) != REG_LIT &&
            AINS_A_EVAL(iins) < 256 &&  AINS_A_EVAL(iins) > 0 ) 
        {
            if( AINS_CODE(iins) == I_addq ||
                AINS_CODE(iins) == I_addl ||
                AINS_CODE(iins) == I_and ||
                AINS_CODE(iins) == I_bis ||
                AINS_CODE(iins) == I_cmpeq ||
                AINS_CODE(iins) == I_eqv ||
                AINS_CODE(iins) == I_mull ||
                AINS_CODE(iins) == I_mulq ||
                AINS_CODE(iins) == I_xor )
            {
                if( verbose )
                {
                    fprintf(fp,"\t[*] switching operands %s %016lx in %d [%s] %ld\n",
                            AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl));
                }
                

                AinsSwitchOperands(iins);
                count++;
                if( ABBL_IS_HOT(ibbl) )
                    counth++;
            }

            if( AINS_CODE(iins) == I_cmple ||
                AINS_CODE(iins) == I_cmplt ||
                AINS_CODE(iins) == I_cmpule ||
                AINS_CODE(iins) == I_cmpult)
            {
                if( AinsTransform(iins,OpTransNegateOpOne,OpTransNegateOpTwo,0) )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] negating %s %016lx in %d [%s] %ld\n",
                                AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
                    }

                    AinsTransform(iins,OpTransNegateOpOne,OpTransNegateOpTwo,1);
                    AinsSwitchOperands(iins);
                    switch( AINS_CODE(iins) )
                    {
                    case I_cmple:
                        AINS_CODE(iins) = I_cmplt;
                        break;
                    case I_cmplt:
                        AINS_CODE(iins) = I_cmple;
                        break;
                    case I_cmpule:
                        AINS_CODE(iins) = I_cmpult;
                        break;
                    case I_cmpult:
                        AINS_CODE(iins) = I_cmpule;
                        break;
                    default:
                        ERROR(("unexpected opcode\n"));
                    }
                    
                    if( verbose )
                    {
                        PrintBblSuper(stdout,ibbl);
                    }

                    count++;
                    if( ABBL_IS_HOT(ibbl) )
                        counth++;
                }

            }
            

            if( AINS_CODE(iins) == I_subq ||
                AINS_CODE(iins) == I_subl )
            {
                if( AinsTransform(iins,OpTransInvertOpOne,OpTransInvertOpTwo,0) )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] inverting %s %016lx in %d [%s] %ld\n",
                                AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
                    }
                                        
                    AinsTransform(iins,OpTransInvertOpOne,OpTransInvertOpTwo,1);
                    AinsSwitchOperands(iins);
                                    
                    if( verbose )
                    {
                        PrintBblSuper(stdout,ibbl);
                    }

                    count++;
                    if( ABBL_IS_HOT(ibbl) )
                        counth++;
                }
                
            }
        }

        if( (AINS_FLAGS(iins) & FI_EVAL_B) && ((INT32)AINS_B_EVAL(iins)) < 0 && ((INT32)AINS_B_EVAL(iins)) > -256)
        {
            if( AINS_CODE(iins) == I_addl )
            {
                HURRAH(("addl -> subl (untested)\n"));
                                
                AINS_CODE(iins) = I_subl;
                AINS_REG_B(iins) = REG_LIT;
                AINS_DATA(iins) =   -((INT32)AINS_B_EVAL(iins));
                AINS_B_EVAL(iins) = -((INT32)AINS_B_EVAL(iins));
                count++;
                if( ABBL_IS_HOT(ibbl) )
                    counth++;

            }
            else if( AINS_CODE(iins) == I_subl )
            {
                HURRAH(("subl -> addl (untested)\n"));
                
                AINS_CODE(iins) = I_addl;
                AINS_REG_B(iins) = REG_LIT;
                AINS_DATA(iins) =   -((INT32)AINS_B_EVAL(iins));
                AINS_B_EVAL(iins) = -((INT32)AINS_B_EVAL(iins));
                count++;
                if( ABBL_IS_HOT(ibbl) )
                    counth++;

            }
        }
    }
    
    
    change = (count)>0;

    InfoStatus(fp,"\t[OptConstuseone%c] changed: %d#%d\n",CHANGE(change),count,counth);

    return change;
}

/*=======================================================
=======================================================*/

static NUMBER TabEqNe[] =
{
    I_beq,I_bne,
    I_bne,I_beq,
    I_cmoveq,I_cmovne,
    I_cmovne,I_cmoveq,
    INVALID,INVALID
};

static NUMBER TabEqEq[] =
{
    I_beq,I_beq,
    I_bne,I_bne,
    I_cmoveq,I_cmoveq,
    I_cmovne,I_cmovne,
    INVALID,INVALID
};

static NUMBER TabEqLt[] =
{
    I_beq,I_blt,
    I_bne,I_bge,
    I_cmoveq,I_cmovlt,
    I_cmovne,I_cmovge,
    INVALID,INVALID
};

static NUMBER TabEqGt[] =
{
    I_beq,I_bgt,
    I_bne,I_ble,
    I_cmoveq,I_cmovgt,
    I_cmovne,I_cmovle,
    INVALID,INVALID
};

static NUMBER TabEqLe[] =
{
    I_beq,I_ble,
    I_bne,I_bgt,
    I_cmoveq,I_cmovle,
    I_cmovne,I_cmovgt,
    INVALID,INVALID
};

static NUMBER TabEqGe[] =
{
    I_beq,I_bge,
    I_bne,I_blt,
    I_cmoveq,I_cmovge,
    I_cmovne,I_cmovlt,
    INVALID,INVALID
};



/*=======================================================
@FUN
@END
=======================================================*/
static BOOL AinsCanTransformAllUses(INDEX iins_the,NUMBER *OpOneTrans, REG the_reg)
{
    INDEX reg;
    INDEX jins;
    BOOL  def;

    def = 0;
    
    reg = AINS_REG_C(iins_the);

    for(jins=AINS_NEXT(iins_the); VALID(jins); jins = AINS_NEXT(jins))
    {
        if( AINS_REG_B(jins) == reg )
                return 0;
        if( AINS_REG_A(jins) == reg )
        {
            if(!VALID(OpcodeTranslate(AINS_CODE(jins),OpOneTrans)) )
                return 0;
            if( def )
                return 0;
        }
        
        if( AINS_REG_C(jins) == reg )
        {
            if( AINS_TYPE(jins) == IT_ICM || AINS_TYPE(jins) == IT_FCM ) return 0;
            else return 1;
        }

        if( AINS_REG_C(jins) == the_reg )
            def = 1;
        
    }
    
    if( ABBL_OUT(AINS_BBL(iins_the)) & REG2BIT(reg) )
        return 0;
    return 1;
}

/*=======================================================
@OPT
Optimize compare (cmpxx) instrutions and instructions that are used
as compare instruction (subq,xor).

Example:
            
@END
=======================================================*/
static NUMBER AinsTransform2(INDEX iins_the,NUMBER *OpOneTrans, REG the_reg )
{
    INDEX reg;
    INDEX jins;
    NUMBER newop;
    NUMBER n = 0;
    BOOL  def;

    def = 0;
    
    reg = AINS_REG_C(iins_the);

    for(jins=AINS_NEXT(iins_the); VALID(jins); jins = AINS_NEXT(jins))
    {
        if( AINS_REG_A(jins) == reg )
        {
            newop = OpcodeTranslate(AINS_CODE(jins),OpOneTrans);
            if( VALID(newop) && !def )
            {
                AINS_CODE(jins) = newop;
                AINS_REG_A(jins) = the_reg;
                n++;
            }
        }
        if( AINS_REG_C(jins) == reg )
            break;
        
        if( AINS_REG_C(jins) == the_reg )
            def = 1;
    }

    return n;
}

/*=======================================================
@OPT
Optimize compare instrutions with zero and their subsequent uses
Example:
            
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstUseTwo"

extern NUMBER OptConstUseTwo(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constusetwo");
    BOOL change;
    BOOL kill;
    NUMBER countc=0;
    NUMBER countch=0;
    NUMBER countk=0;
    NUMBER countkh=0;
    
    NUMBER n;
    
    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    
    FOREACH_LIVE_FUN(ifun)
    FOREACH_FUN_BBL(ifun,ibbl)
    FOREACH_BBL_INS(ibbl,iins)
        {

            if( ( AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
                continue;

            if( AINS_REG_A(iins) == AINS_REG_B(iins) )
            {
                switch(AINS_CODE(iins))
                {
                case I_cmpeq:
                case I_cmple:
                case I_cmpule:
                    AINS_FLAGS(iins) |= FI_EVAL_C;
                    AINS_C_EVAL(iins) = 1;
                    countc++;
                    if( ABBL_IS_HOT(ibbl) )
                        countch++;
                    if( verbose )
                        fprintf(fp,"\t[*] Evaluated compare with identical operands %d %d\n",iins,ibbl);
                    break;
                case I_cmpult:
                case I_cmplt:
                    AINS_FLAGS(iins) |= FI_EVAL_C;
                    AINS_C_EVAL(iins) = 0;
                    countc++;
                    if( ABBL_IS_HOT(ibbl) )
                        countch++;
                    if( verbose )
                        fprintf(fp,"\t[*] Evaluated compare with identical operands %d %d\n",iins,ibbl);
                    break;
                }
            }

            
            if( AINS_CODE(iins)==I_cmplt && (AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins)==1 )
            {
                AINS_REG_B(iins) = REG_ZERO;
                AINS_B_EVAL(iins) = 0;
                AINS_CODE(iins) = I_cmple;
                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
                if( verbose )
                    fprintf(fp,"\t[*] Simplified compare with one %d %d\n",iins,ibbl);
            }

            if( AINS_CODE(iins)==I_cmpult && (AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins)==1 )
            {
                AINS_REG_B(iins) = REG_ZERO;
                AINS_B_EVAL(iins) = 0;
                AINS_CODE(iins) = I_cmpule;
                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
                if( verbose )
                    fprintf(fp,"\t[*] Simplified compare with one %d %d\n",iins,ibbl);
            }

            if( AINS_CODE(iins)==I_cmpeq && AINS_REG_A(iins)==REG_ZERO ) 
            {
                kill =  AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_B(iins));
                if( AINS_REG_B(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                        PrintBblSuper(stdout,ibbl);
                                    
                n = AinsTransform2(iins,TabEqNe,AINS_REG_B(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }

#if 1
            if( AINS_CODE(iins)==I_lda && AINS_DATA(iins) < 0 && AINS_DATA(iins) >= -255 )
            {
                if( AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_C(iins)) )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                    
                    AinsMakeCmpeq(iins,AINS_REG_B(iins),REG_LIT, AINS_REG_C(iins),-AINS_DATA(iins));
                    n=AinsTransform2(iins,TabEqNe,AINS_REG_C(iins));
                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;

                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }
                        
            if( AINS_CODE(iins)==I_subq )
            {
                if( AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_C(iins)) )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                                        
                    AinsMakeCmpeq(iins,AINS_REG_A(iins),AINS_REG_B(iins), AINS_REG_C(iins),AINS_DATA(iins));
                    n=AinsTransform2(iins,TabEqNe,AINS_REG_C(iins));
                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;

                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }

            if( AINS_CODE(iins)==I_xor )
            {
                if( AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_C(iins)) )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                    AinsMakeCmpeq(iins,AINS_REG_A(iins),AINS_REG_B(iins), AINS_REG_C(iins),AINS_DATA(iins));
                    n=AinsTransform2(iins,TabEqNe,AINS_REG_C(iins));
                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }
#endif
            
            if( AINS_CODE(iins)==I_cmpeq && AINS_REG_B(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                                    
                n=AinsTransform2(iins,TabEqNe,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            
            if( AINS_CODE(iins)==I_cmple && AINS_REG_A(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqLt,AINS_REG_B(iins));
                if( AINS_REG_B(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);

                n=AinsTransform2(iins,TabEqLt,AINS_REG_B(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            

            if( AINS_CODE(iins)==I_cmple && AINS_REG_B(iins)==REG_ZERO )
            {
                kill =  AinsCanTransformAllUses(iins,TabEqGt,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                n=AinsTransform2(iins,TabEqGt,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }

                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }


            if( AINS_CODE(iins)==I_cmplt && AINS_REG_A(iins)==REG_ZERO )
            {
                kill =  AinsCanTransformAllUses(iins,TabEqLe,AINS_REG_B(iins));
                if( AINS_REG_B(iins) != AINS_REG_C(iins) || kill )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                    n=AinsTransform2(iins,TabEqLe,AINS_REG_B(iins));
                    if( kill )
                    {
                        countk++;
                        if( ABBL_IS_HOT(ibbl) )
                            countkh++;
                        AinsKill(iins);
                    }

                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }
            
            if( AINS_CODE(iins)==I_cmplt && AINS_REG_B(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqGe,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                n=AinsTransform2(iins,TabEqGe,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }

                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            
            if( AINS_CODE(iins)==I_cmpult && AINS_REG_A(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqEq,AINS_REG_B(iins));
                if( AINS_REG_B(iins) == AINS_REG_C(iins) && !kill ) continue;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                n = AinsTransform2(iins,TabEqEq,AINS_REG_B(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }

                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            
            if( AINS_CODE(iins)==I_cmpult && AINS_REG_B(iins)==REG_ZERO )
            {
                AINS_FLAGS(iins) |= FI_EVAL_C;
                AINS_C_EVAL(iins) = 0;

                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
            }


            if( AINS_CODE(iins)==I_cmpule && AINS_REG_A(iins)==REG_ZERO )
            {
                AINS_FLAGS(iins) |= FI_EVAL_C;
                AINS_C_EVAL(iins) = 1;

                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
            }

            /* same as I_cmpeq */
            
            if( AINS_CODE(iins)==I_cmpule && AINS_REG_B(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;
                if( verbose )
                {
                    PrintIns(stdout,iins);
                    PrintBblSuper(stdout,ibbl);
                }
                
                n = AinsTransform2(iins,TabEqNe,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
        }
    
    change = (countc+countk)>0;

    InfoStatus(fp,"\t[OptConstusetwo%c] changed: %d#%d killed: %d#%d\n",CHANGE(change),countc,countch,countk,countkh);
        
    return change;
}



/*=======================================================
@OPT
Peephole constant uses
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstUseThree"

extern NUMBER OptConstUseThree(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constusethree");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;
    
    INDEX ibbl;
    INDEX iins;

    FOREACH_LIVE_BBL(ibbl)
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_FLAGS(iins) & FI_EVAL_A )
            {
                switch( AINS_CODE(iins) )
                {
                case I_and:
                    if( AINS_A_EVAL(iins) >= 256 && VALID(BinaryCondenseOctets(AINS_A_EVAL(iins))) )
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] Transforming: %d\n",iins);
                            PrintBblSuper(fp,ibbl);
                        }
                        AinsMakeZapnot(iins,AINS_REG_B(iins),REG_LIT,AINS_REG_C(iins),BinaryCondenseOctets(AINS_A_EVAL(iins)));
                        if( verbose )
                        {
                            PrintIns(fp,iins);
                        }
                        count++;
                        if( ABBL_IS_HOT(ibbl) ) counth++;
                    }
                    break;
                }
            }

            if( AINS_FLAGS(iins) & FI_EVAL_B )
            {
                switch( AINS_CODE(iins) )
                {
                case I_and:
                    if( AINS_B_EVAL(iins) >= 256 && VALID(BinaryCondenseOctets(AINS_B_EVAL(iins))) )
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] Transforming: %d\n",iins);
                            PrintBblSuper(fp,ibbl);

                        }
                        
                        AinsMakeZapnot(iins,AINS_REG_A(iins),REG_LIT,AINS_REG_C(iins),BinaryCondenseOctets(AINS_B_EVAL(iins)));
                        if( verbose )
                        {
                            PrintIns(fp,iins);
                        }
                        count++;
                        if( ABBL_IS_HOT(ibbl) ) counth++;
                    }
                    break;
                }
            }
        }

    change = (count)>0;

    InfoStatus(fp,"\t[OptConstusethree%c] changed: %d#%d\n",CHANGE(change),count,counth);
        
    return change;
}


/*=======================================================
=======================================================*/



