/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.dupbbl.c,v 1.16 1999/06/15 21:35:52 muth Exp $
  Comment:  This file contains the code for the bbl duplication optimizations 
                
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
@OPT
Duplicate small basic blocks that are reached via a ET_NORM edge
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptDupBblOne"

extern BOOL OptDupBblOne(FILE *fp, BOOL no_code_growth)
{
    BOOL change = 0;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Dupbblone");
    INDEX jbbl;
    INDEX ibbl;
    INDEX pred;
    INDEX succ;
    INDEX iins;
    INDEX jins;
    NUMBER bcount = 0;
    NUMBER bcounth = 0;
    NUMBER icount = 0;
    NUMBER icounth = 0;
    NUMBER n;
    
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( no_code_growth ) continue;
        
        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;

        /* end of temporary thing */

#if 1
        if( ABBL_TYPE(ibbl) == BT_CALL ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_PAL ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_SWITCH ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_EXIT ) continue;   /* do not duplicate calls */
#else
        if( ABBL_TYPE(ibbl) != BT_NORM && ABBL_TYPE(ibbl) != BT_RET ) continue;   /* do not duplicate calls */
#endif
        
        FOREACH_BBL_PRED(ibbl,pred)
        {
            
            if( AEDG_TYPE(pred) != ET_NORM ) continue;
            jbbl = AEDG_HEAD(pred);
            if( jbbl == ibbl )
            {
                WARNING(("found self loop normal edge\n"));
                continue;
            }

            n = AbblNInss(ibbl);
            if( n > 4 || n < 1 ) continue;
            
            if( AEDG_IS_INTER(pred) )
            {
                INDEX ifun = ABBL_FUN(AEDG_HEAD(pred));
                INDEX jfun = ABBL_FUN(AEDG_TAIL(pred));
                
                WARNING(("Merging bbls across function boundaries %d [%s] -> %d [%s]\n",ifun,AFUN_NAME(ifun),jfun,AFUN_NAME(jfun)));
#if 0
                PrintEdgSuper(stdout,pred);
#endif
            }
            
            if( verbose )
            {
                fprintf(fp,"\t[*] merging edg %d: hd %d tl %d iins %d\n",pred,jbbl,ibbl,AbblNInss(ibbl));
                /*
                PrintBblSuper(stdout,jbbl);
                PrintBblSuper(stdout,ibbl);
                */
            }

            icount += n;
            if( ABBL_IS_HOT(ibbl) ) icounth += n;
            
            
            bcount++;
            if( ABBL_IS_HOT(jbbl) )
                bcounth++;
            
            AedgKillFancy(pred);
            ABBL_COUNT(ibbl) -= ABBL_COUNT(jbbl);
            if( ABBL_COUNT(ibbl) < 0 )
                ABBL_COUNT(ibbl) = 0;
            
            ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
            
            FOREACH_BBL_INS(ibbl,iins)
            {
                jins = AinsDup(iins);
                AINS_BBL(jins) = jbbl;
                AinsAppend(jins);
            }
        
            FOREACH_BBL_SUCC(ibbl,succ)
            {
                AedgLinkFancy(AedgNewFancy(jbbl,AEDG_TAIL(succ),AEDG_TYPE(succ)));
            }
            /*
            PrintBblSuper(stdout,jbbl);
            */
        }
        
    }
        
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (bcount>0);

    InfoStatus(fp,"\t[OptDupbblone%c] Dup Bbls: %d#%d  Inss: %d#%d {%d ms}\n",
            CHANGE(change),bcount,bcounth,icount,icounth,time);
    
    return change;
}

/*=======================================================
@OPT
Duplicate basic blocks with a conditional branch if this
will help evaluating the branch
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptDupBblTwo"

extern BOOL OptDupBblTwo(FILE *fp, BOOL no_code_growth)
{
    BOOL change = 0;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Dupbbltwo");
    INDEX jbbl;
    INDEX ibbl;
    INDEX xbbl;
    INDEX pred;
    INDEX succ;
    INDEX iins;
    REG reg;
    
    NUMBER bcount = 0;
    NUMBER bcounth = 0;
    NUMBER icount = 0;
    NUMBER icounth = 0;

    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( no_code_growth ) continue;

        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;

        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;

        if( ABBL_TYPE(ibbl) == BT_CALL ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_PAL ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_SWITCH ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_EXIT ) continue;   /* do not duplicate calls */

        if( AbblNpreds(ibbl) <= 1 ) continue;  /* let OptMerge do the work */

        /*
          make sure that value is not defined in current bbl which is also the extended bbl
          since it has multiple predecessors
           */

        iins = ABBL_INS_LAST(ibbl);
        reg = AINS_REG_A(iins);
        if( VALID(AinsFindDefInExtBbl(iins, reg)) ) continue;
        
        FOREACH_BBL_PRED(ibbl,pred)
        {
            if( AEDG_TYPE(pred) != ET_TRUE &&
                AEDG_TYPE(pred) != ET_FALSE &&
                AEDG_TYPE(pred) != ET_NORM  ) continue;

            jbbl = AEDG_HEAD(pred);

            iins = ABBL_INS_LAST(jbbl);
            iins = AinsFindDefInExtBbl(iins, reg);
            if( !(AINS_FLAGS(iins) & FI_EVAL_C) ) continue;
                
            if( AEDG_IS_INTER(pred) )
            {
                WARNING(("Merging bbls across function boundaries\n"));
                PrintEdgSuper(fp,pred);
            }
                        
            if( verbose )
            {
                fprintf(fp,"\tmerging edg %d: %d %d\n",pred,jbbl,ibbl);
                PrintBblSuper(fp,jbbl);
                PrintBblSuper(fp,ibbl);
            }

            bcount++;
            if( ABBL_IS_HOT(ibbl) )
                bcounth++;
            
            AedgKillFancy(pred);
            ABBL_COUNT(ibbl) -= AEDG_COUNT(pred);
            if( ABBL_COUNT(ibbl) < 0 )
                ABBL_COUNT(ibbl) = 0;
            
            if( AEDG_TYPE(pred) != ET_NORM )
            {
                xbbl = AbblNew();
                ABBL_FUN(xbbl) = ABBL_FUN(jbbl);
                ABBL_COUNT(xbbl) = AEDG_COUNT(pred);
                if( ABBL_COUNT(xbbl) < 0 )
                    ABBL_COUNT(xbbl) = 0;
                AedgLinkFancy( AedgNewFancy(jbbl,xbbl,AEDG_TYPE(pred)));
            }
            else
            {
                xbbl = jbbl;
            }
            
        
            ABBL_TYPE(xbbl) = ABBL_TYPE(ibbl);
            
            FOREACH_BBL_INS(ibbl,iins)
            {
                    INDEX jins = AinsDup(iins);
                    AINS_BBL(jins) = xbbl;
                    AinsAppend(jins);
                    icount++;
                    if( ABBL_IS_HOT(ibbl) )
                        icounth++;
            }
                            
            FOREACH_BBL_SUCC(ibbl,succ)
            {
                INDEX jedg = AedgNewFancy(xbbl,AEDG_TAIL(succ),AEDG_TYPE(succ));
                AedgLinkFancy(jedg);
            }

            if( xbbl !=jbbl )
                AbblAppendFancy(xbbl);
        }
    }
        
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (bcount>0);

    InfoStatus(fp,"\t[OptDupbbltwo%c] Dup Bbls: %d#%d  Inss: %d#%d {%d ms}\n",
            CHANGE(change),bcount,bcounth,icount,icounth,time);
    
    return change;
}

/*=======================================================
@OPT
Duplicate basic blocks which constitue a single bbl loop
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptkDupBblThree"

#define BBL_MAX_BEFORE_LENGTH 32
#define BBL_MAX_AFTER_LENGTH 64

extern BOOL OptDupBblThree(FILE *fp, BOOL no_code_growth)
{
    BOOL change = 0;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Dupbblthree");
    INDEX ibbl;
    INDEX xbbl;
    INDEX last;
    INDEX iedg;
    NUMBER copies;
    
    NUMBER bcount = 0;
    NUMBER bcounth = 0;
    NUMBER icount = 0;
    NUMBER icounth = 0;

    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( no_code_growth ) continue;

        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;

        if( !ABBL_IS_HOT(ibbl) ) continue;
        
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;

        if( ABBL_TYPE(ibbl) == BT_CALL ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_PAL ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_SWITCH ) continue;   /* do not duplicate calls */
        if( ABBL_TYPE(ibbl) == BT_EXIT ) continue;   /* do not duplicate calls */

        if( AbblNpreds(ibbl) <= 1 ) continue;  /* let OptMerge do the work */

        if( AEDG_TAIL(ABBL_EDG_TRUE(ibbl)) == ibbl )
        {
            AbblBranchTwist(ibbl);
        }
        
        if( AEDG_TAIL(ABBL_EDG_FALSE(ibbl)) != ibbl ) continue;

        /* this is just a hack until I come up with a cache size based heuristic */
           
        if( AbblNInss(ibbl) >= BBL_MAX_BEFORE_LENGTH ) continue;
        copies = BBL_MAX_AFTER_LENGTH / (AbblNInss(ibbl)+1);
        ABBL_COUNT(ibbl) /= copies;
        copies--;

        last = ibbl;
        
        while( copies-- )
        {
            if( verbose )
            {
                fprintf(fp,"\t[*] duplicating bbl %d\n",last);
                PrintBblSuper(fp,last);
            }

            bcount++;
            if( ABBL_IS_HOT(last) )
                bcounth++;

            icount += AbblNInss(last);
            if( ABBL_IS_HOT(last) )
                icounth += AbblNInss(last);
                        
            xbbl = AbblDup(last);   
            /* copies count field */
            ABBL_FUN(xbbl) = ABBL_FUN(last);

            
            FOREACH_BBL_SUCC(last,iedg)
            {
                INDEX jedg = AedgNewFancy(xbbl,AEDG_TAIL(iedg),AEDG_TYPE(iedg));
                AedgLinkFancy(jedg);
            }

            AbblAppendFancy(xbbl);

            AedgKillFancy( ABBL_EDG_FALSE(last) );
            AedgLinkFancy( AedgNewFancy(last,xbbl,ET_FALSE) );

            last = xbbl;
        }
    }
        
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (bcount>0);

    InfoStatus(fp,"\t[OptDupbblthree%c] Dup Bbls: %d#%d  Inss: %d#%d {%d ms}\n",
            CHANGE(change),bcount,bcounth,icount,icounth,time);
    
    return change;
}

/*=======================================================
=======================================================*/











