/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.propagate.c,v 1.30 1999/05/31 03:33:45 muth Exp $
  Comment:  This file contains the optimization code for
            Copy Propagation
            Move Elimination
            Reload Avoidance
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

#define AINS_IS_INT_MOVE(iins) (AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0 && AINS_REG_B(iins) != REG_ZERO)
#define AINS_IS_FLT_MOVE(iins) (AINS_CODE(iins) == I_cpys && AINS_REG_A(iins) == AINS_REG_B(iins) && AINS_REG_B(iins) != REG_FZERO)


/*=======================================================
@FUN
Return the lowest positon of a set bit in rl.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "RegsGetFirst"

extern REG RegsGetFirst(REGLIST rl)
{
    REG reg = 0;
    
    if( rl == 0 ) return INVALID;
    while( !(rl & 1) )
    {
        rl>>=1;
        reg++;
    }
    return reg;
}

/*=======================================================
@OPT
For each move instruction which moves ra to rb:
Try to replace as many uses of rb with those of rb
This can even handle moves with an offset, eg.lda's
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCopyPropObsolete"

extern BOOL OptCopyProp(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Copyprop");
    BOOL   change = 0;
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    INDEX  ibbl;
    INDEX iins;
    INDEX jins;
    REG source;
    REG target;
    
    clock_t time = clock();
    
     
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_CODE(iins) != I_lda && !AINS_IS_FLT_MOVE(iins) )  
                continue;

            source = AINS_REG_B(iins);
            target = AINS_REG_C(iins);

            if( source == target ) continue;
            if( source == REG_ZERO || source == REG_FZERO ) continue;
            if( target == REG_ZERO || target == REG_FZERO ) continue;
            
            for(jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins) )
            {
                if( AINS_REG_A(jins) == target && AINS_TYPE(iins) != IT_CST ) 
                {
                    if( AINS_CODE(iins) == I_cpys )
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                            PrintBblSuper(fp,ibbl);
                        }
                        
                        AINS_REG_A(jins) = source;
                        pcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;                                                                            

                    }
                    else if( AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0)
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                            PrintBblSuper(fp,ibbl);
                        }

                        AINS_REG_A(jins) = source;
                        pcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;                            
                    }
                }

                if( AINS_REG_B(jins) == target )
                {
                    if( AINS_CODE(iins) == I_cpys )
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                            PrintBblSuper(fp,ibbl);
                        }

                        AINS_REG_B(jins) = source;
                        pcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;
                    }
                    else if( AINS_CODE(iins) == I_lda)
                    {
                        
                        if( AINS_DATA(iins) == 0)
                        {
                            if( verbose )
                            {
                                fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                                PrintBblSuper(fp,ibbl);
                            }

                            AINS_REG_B(jins) = source;
                            pcount++;
                            if( ABBL_IS_HOT(ibbl) )
                                pcounth++;                            
                        }
                        else if( AINS_CODE(jins) == I_lda ||
                                 AINS_TYPE(jins) == IT_ILD ||
                                 AINS_TYPE(jins) == IT_FLD ||
                                 AINS_TYPE(jins) == IT_ULD ||
                                 AINS_TYPE(jins) == IT_LLD ||
                                 AINS_TYPE(jins) == IT_IST ||
                                 AINS_TYPE(jins) == IT_FST ||
                                 AINS_TYPE(jins) == IT_UST ||
                                 AINS_TYPE(jins) == IT_CST )
                        {
                           if( AINS_DATA(jins) + AINS_DATA(iins) < (1<<15) &&
                               AINS_DATA(jins) + AINS_DATA(iins) >= -(1<<15) )
                           {
                               if( verbose )
                               {
                                   fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                                   PrintBblSuper(fp,ibbl);
                               }
                               
                               AINS_REG_B(jins) = source;
                               AINS_DATA(jins) += AINS_DATA(iins);
                               if( AINS_FLAGS(jins) & FI_EVAL_B )
                               {
                                   AINS_B_EVAL(jins) -= AINS_DATA(iins);
                               }
                               pcount++;
                               if( ABBL_IS_HOT(ibbl) )
                                   pcounth++;
                                                                  
                               if( verbose )
                               {
                                   PrintBblSuper(fp,ibbl);
                               }
                           }
                        }
                    }
                }

                if( AINS_REG_C(jins) == target )
                    break;

                if( AINS_REG_C(jins) == source )
                    break;
            }
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = pcount>0;

    InfoStatus(fp,"\t[OptCopyprop%c] Propagated Values: %d#%d {%d ms}\n",CHANGE(change),pcount,pcounth,time);

    return change;
}

/*=======================================================
@OPT
Eliminate Move "move src,dst" by changing all subsequent uses of dst into src.
Of course, src must be available at the last use of dst.
The whole optimization is complicated my conditional move instructions.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptMoveElimOne"

extern BOOL OptMoveElimOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Moveelimone");
    BOOL   change = 0;
    
    NUMBER kcount = 0;
    NUMBER kcounth = 0;
    INT64  weight = 0;
    clock_t time = clock();

    INDEX  ibbl;
    REG    reg_source;
    REG    reg_target;

    INDEX  iins;

    INDEX  target_def;   /* the move instruction */
    INDEX  target_last_use;
    INDEX  source_last_use;


    //    verbose = 1;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,target_def)
        {
            if( AINS_CODE(target_def) != I_lda && !AINS_IS_FLT_MOVE(target_def) )  
                continue;

            reg_source = AINS_REG_B(target_def);
            reg_target = AINS_REG_C(target_def);

#if 0	
	if( reg_source == reg_target && reg_source == 9)
	{
		printf("QQQQQQQQQQQQQQQQQQQQQQQQQQQQQQQ\n");
		PrintIns(stdout,target_def);
		PrintBblSuper(stdout,ibbl);
	}
	else
		continue;
#endif	
		
            /* compute the last use of source and target.
               Note, that the Find..() functions treat conditional moves special */
           
            source_last_use = FindInstructionRegLastUse(target_def, reg_source);
            target_last_use = FindInstructionRegLastUse(target_def, reg_target);

            /* we cannot do anything if the last use of the target is outside of the bbl */
#if 0
            printf("last use: %d\n",target_last_use);
#endif
            if( !VALID(target_last_use) ) continue;

	    
            /* dead code */
            
            if( target_last_use == target_def ) continue;

            /* do not touch REG_ZERO, placing it into AINS_REG_C causes incorrectness !!! */

            if( reg_source == REG_ZERO ) continue;
            
               
            /* between (target_def and source_last_use) and (target_def and target_last)
               there might be conditional moves defining reg_source and reg_target
               if such a conditional move occurs before either source_last_use
               or target_last_use we cannot do the renaming
               */
               
            for( iins = target_def; VALID(iins); iins = AINS_NEXT(iins) )
            {
                if( iins == source_last_use || iins == target_last_use )
                {
                    iins = INVALID;
                    break;
                }

                if( AINS_TYPE(iins) != IT_ICM && AINS_TYPE(iins) != IT_FCM && AINS_TYPE(iins) != IT_CST ) continue;
                if( AINS_REG_C(iins) == reg_target || AINS_REG_C(iins) == reg_source )
                    break;
            }

            if( VALID(iins) ) continue;
#if 0
	    printf("past check 1\n");
#endif       
            /*
              check whether source reg is still available at target_last_use
              this will always be the case if the source range extends to past the target range!
             */
            
            for( iins = AINS_NEXT(target_def); iins != target_last_use; iins = AINS_NEXT(iins) )
            {
            	
                if( AINS_REG_C(iins) == reg_source )
                {
                    iins = INVALID;
                    break;
                }
                    
            }

            if( !VALID(iins) ) continue;
#if 0
               	    printf("past check 2\n");
#endif    
            /*
              if the move is and lda with nozero offset
              check whether target register is only used with load/store type instructions
             */

            if( AINS_CODE(target_def) == I_lda && AINS_DATA(target_def) != 0 )
            {
                for( iins = AINS_NEXT(target_def); VALID(iins); iins = AINS_NEXT(iins) )
                {
                    if( !(AinsUse(iins) & REG2BIT(reg_target)) ) continue;
                    
                    if( AINS_REG_A(iins) == reg_target )  break;
                    
                    if( AINS_CODE(iins) == I_lda ||
                        AINS_TYPE(iins) == IT_ILD ||
                        AINS_TYPE(iins) == IT_FLD ||
                        AINS_TYPE(iins) == IT_ULD ||
                        AINS_TYPE(iins) == IT_LLD ||
                        AINS_TYPE(iins) == IT_IST ||
                        AINS_TYPE(iins) == IT_FST ||
                        AINS_TYPE(iins) == IT_UST ||
                        AINS_TYPE(iins) == IT_CST )
                    {
                        if( AINS_DATA(iins) + AINS_DATA(target_def) >= (1<<15) ) break;
                        if( AINS_DATA(iins) + AINS_DATA(target_def) < -(1<<15) ) break;
                    }
                    else
                    {
                 	   break;
                    }
                    
                    if( iins == target_last_use )
                    {
                    	iins = INVALID;
                    	break;
                    }
                }

#if 0	
		if ( VALID(iins) )
		{
			printf("bad instruction\n");
			PrintIns(stdout,iins);
		}
#endif
                if( VALID(iins) ) continue;
            }
            
#if 0
               printf("past check 3\n");
#endif 
            
            
            if( verbose )
            {
                
                fprintf(fp,"\t[*] Removing move instruction: %d lastuse: %d (renaming %d -> %d)\n",
                        target_def,target_last_use,reg_target,reg_source);
                PrintBblSuper(fp,ibbl);
            }

            
            for(iins = AINS_NEXT(target_def); VALID(iins); iins = AINS_NEXT(iins) )
            {
                if( AINS_REG_A(iins) == reg_target )
                {
#if 0
                    printf("found A use in %d\n",iins);
#endif               
                    if( AINS_CODE(target_def) == I_cpys )
                    {
                        AINS_REG_A(iins) = reg_source;

                    }
                    else 
                    {
                        CASSERT(  AINS_CODE(target_def) == I_lda,("bad target code: %s\n",AINS_CODE(target_def)));
                        CASSERT( AINS_DATA(target_def) == 0,("unexpected use of reg %d in ins %d\n",reg_target,iins));
                        AINS_REG_A(iins) = reg_source;
                    }
                }

                if( AINS_REG_B(iins) == reg_target )
                {
#if 0
                    printf("found B use in %d\n",iins);
#endif               
                    if( AINS_CODE(target_def) == I_cpys )
                    {
                        AINS_REG_B(iins) = reg_source;
                    }
                    else 
                    {
                        CASSERT(  AINS_CODE(target_def) == I_lda,("bad target code: %s\n",AINS_NAME(target_def)));
                        
                        if( AINS_DATA(target_def) == 0)
                        {
                            AINS_REG_B(iins) = reg_source;
                        }
                        else
                        {
                            CASSERT( AINS_CODE(iins) == I_lda ||
                                    AINS_TYPE(iins) == IT_ILD ||
                                    AINS_TYPE(iins) == IT_FLD ||
                                    AINS_TYPE(iins) == IT_ULD ||
                                    AINS_TYPE(iins) == IT_LLD ||
                                    AINS_TYPE(iins) == IT_IST ||
                                    AINS_TYPE(iins) == IT_FST ||
                                    AINS_TYPE(iins) == IT_UST ||
                                    AINS_TYPE(iins) == IT_CST,("unexpected ins %s\n",AINS_NAME(iins)));
                            
                            CASSERT( AINS_DATA(iins) + AINS_DATA(target_def) < (1<<15) &&
                                    AINS_DATA(iins) + AINS_DATA(target_def) >= -(1<<15),("unexpected offset\n"));
                            
                            AINS_REG_B(iins) = reg_source;
                            AINS_DATA(iins) += AINS_DATA(target_def);

                            if( AINS_FLAGS(iins) & FI_EVAL_B )
                            {
                                   AINS_B_EVAL(iins) -= AINS_DATA(target_def);
                            }
                        }
                    }
                }

                if( AINS_REG_C(iins) == reg_target )
                {
#if 0
                    printf("found C use in %d\n",iins);
#endif                                   
                    if( AINS_TYPE(iins) == IT_ICM || AINS_TYPE(iins) == IT_FCM || AINS_TYPE(iins) == IT_CST )
                    {
                        if( AINS_CODE(target_def) == I_cpys )
                        {
                            AINS_REG_C(iins) = reg_source;
                        }
                        else if( AINS_CODE(target_def) == I_lda )
                        {
                            CASSERT( AINS_DATA(target_def) == 0,("weird\n"));
                            AINS_REG_C(iins) = reg_source;
                        }
                    }
                    else
                    {
                        break;
                    }
                    
                   
                }
            }
            

            AinsKill(target_def);
            weight += ABBL_COUNT(ibbl);
            kcount++;
            if( ABBL_IS_HOT(ibbl) )
                kcounth++;

            if( verbose )
            {
                PrintBblSuper(fp,ibbl);
            }
        }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = kcount>0;

    InfoStatus(fp,"\t[OptMoveelimone%c] Eliminated Moves: %d#%d wt: %ld  {%d ms}\n",
            CHANGE(change),kcount,kcounth,weight,time);

    return change;
}

/*=======================================================
=======================================================*/
      
/*=======================================================
@OPT
Eliminate Moves

This is a replacement for the copy propagator and the move eliminator
This works only at the bbl level at the moment.

Given a move "move ra,rb" we want to eliminate it!

We proceed as follows

o find the defintion of ra "ra_def" and the last use of ra "ra_last_use"
o the move instruction is the definition of rb "rb_def" find the last
  use of rb "rb_last_use"

  if: rb_last_use AND ra_last_use lie outside of the bbl we cannot eliminate the move! 

  else if: rb_last_use comes before ra_last_use we just rename rb to ra
     this also applied if rb_last_use is outside of the bbl

  else if: ra_last_use comes before rb_last_use we can try to change ra_def to compute
           the values into a different register if rb_last_use is within the bbl
           we cab pick an arbitrary register with the right lifespan.
           if rb_last_use is outside of the bbl we are limited ti rb as our renaming choice.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptMoveElim"

extern BOOL OptMoveElim(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Moveelim");
    BOOL   change = 0;
    
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    NUMBER kcount = 0;
    NUMBER kcounth = 0;
    INT64  weight = 0;
    clock_t time = clock();

    INDEX  ibbl;
    REG    reg_source;
    REG    reg_target;

    INDEX  iins;
    INDEX  source_def;
    INDEX  target_def;
    INDEX  target_last_use;
    INDEX  source_last_use;
    REGLIST avail;
    REG     bypass;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,target_def)
        {
            if( !AINS_IS_INT_MOVE(target_def) && !AINS_IS_FLT_MOVE(target_def) ) continue;

            reg_source = AINS_REG_B(target_def);
            reg_target = AINS_REG_C(target_def);

            if( reg_target == reg_source ) continue;
            
            source_def = FindInstructionRegDef(target_def, reg_source);
            
                         
            /* source defintion might be outside of the bbl*/

            source_last_use = FindInstructionRegLastUse(target_def, reg_source);
            target_last_use = FindInstructionRegLastUse(target_def, reg_target);

            if( !VALID(target_last_use) && !VALID(source_last_use) ) continue;

            for( iins = target_def; VALID(iins); iins = AINS_NEXT(iins) )
            {
                if( iins == source_last_use || iins == target_last_use )
                {
                    iins = INVALID;
                    break;
                }

                if( AINS_TYPE(iins) != IT_ICM && AINS_TYPE(iins) != IT_FCM && AINS_TYPE(iins) != IT_CST ) continue;
                if( AINS_REG_C(iins) == reg_target || AINS_REG_C(iins) == reg_source )
                    break;
            }
            if( VALID(iins) ) continue;
            
            pcount++;
            if( ABBL_IS_HOT(ibbl) )
                pcounth++;

#if 0
            printf("bbl: %d source: def %d  use %d   target: def %d  use %d\n",ibbl,source_def,source_last_use,target_def,target_last_use);
#endif
            /*
              check whether source reg is still available at target_last_use
              this will always be the case if the source range extends to past the target range!
             */
            
            if( VALID(target_last_use) )
            {
                for( iins = target_def; iins != target_last_use; iins = AINS_NEXT(iins) )
                {
                    if( AINS_REG_C(iins) == reg_source )
                    {
                        iins = INVALID;
                        break;
                    }
                    
                }

                if( VALID(iins) )
                {
                    if( verbose )
                    {
                        
                        fprintf(fp,"\t[*] extending the source range over target range for instr: %d\n",target_def);
                        fprintf(fp,"\t[*]bbl: %d source def %d  use %d   target def %d  use %d\n",
                                ibbl,source_def,source_last_use,target_def,target_last_use);
                        PrintBblSuper(fp,ibbl);
                    }
                    RenameRegister(target_def,reg_source);
                    AinsKill(target_def);
                    weight += ABBL_COUNT(ibbl);
                    kcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        kcounth++;
                    
                    if( verbose )
                    {
                        PrintBblSuper(fp,ibbl);
                    }
                    continue;
                }
                
            }

            /*
              last use of target before last use of source
              source range does not extend over target range
             */
            
            if( VALID(source_def) && VALID(source_last_use) )
            {
                avail = MyFindScratchRegs(source_def, target_def);
                avail &=  REG_IS_INT(reg_source) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

                if( REG2BIT(reg_target) & avail )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] extending target range over source range  instr: %d\n",target_def);
                        fprintf(fp,"\t[*]bbl: %d source def %d  use %d   target def %d  use %d\n",
                                ibbl,source_def,source_last_use,target_def,target_last_use);
                        PrintBblSuper(fp,ibbl);
                    }
                    RenameRegister(source_def,reg_target);
                    AinsKill(target_def);
                    weight += ABBL_COUNT(ibbl);
                    kcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        kcounth++;

                    if( verbose )
                    {
                        PrintBblSuper(fp,ibbl);
                    }
                    
                    continue;
                }
                                

                if( VALID(target_last_use)  )
                {
                    avail = MyFindScratchRegs(source_def, target_last_use);
                    avail &=  REG_IS_INT(reg_source) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

                    if( avail )
                    {
                        bypass = RegsGetFirst(avail);

                        if( verbose )
                        {
                            fprintf(fp,"\t[*] creating new bypass range with reg %d for  instr: %d\n",bypass,target_def);
                            fprintf(fp,"\t[*]bbl: %d source def %d  use %d   target def %d  use %d\n",
                                    ibbl,source_def,source_last_use,target_def,target_last_use);
                            PrintBblSuper(fp,ibbl);
                        }
                        
                        RenameRegister(source_def,bypass);
                        RenameRegister(target_def,bypass);
                        AinsKill(target_def);
                        weight += ABBL_COUNT(ibbl);              
                        kcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            kcounth++;
                        
                        if( verbose )
                        {
                            PrintBblSuper(fp,ibbl);
                        }
                        
                        continue;
                    }
                    
                }
            }
        }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = kcount>0;

    InfoStatus(fp,"\t[OptMoveelim%c] Eliminated Moves: %d#%d wt: %ld  Potential: %d#%d{%d ms}\n",
            CHANGE(change),kcount,kcounth,weight,pcount,pcounth,time);

    return change;
}

/*=======================================================
=======================================================*/

extern BOOL AinsDoesOperandSignExtension(INDEX iins)
{
    switch( AINS_CODE(iins) )
    {
    case I_stw:
    case I_stb:
    case I_stl:
    case I_addl:
    case I_s4addl:
    case I_subl:
    case I_s4subl:
    case I_s8addl:
    case I_s8subl:
    case I_addl_v:
    case I_subl_v:
    case I_mull:
    case I_mull_v:
        return 1;

    case I_zap:
        if( (AINS_FLAGS(iins) & FI_EVAL_B) && (AINS_B_EVAL(iins) & 0xf0) == 0xf0 )
            return 1;
        else
            return 0;

    case I_zapnot:
        if( (AINS_FLAGS(iins) & FI_EVAL_B) && (AINS_B_EVAL(iins) | 0xf) == 0xf )
            return 1;
        else
            return 0;
        
    case I_and:
        if( (AINS_FLAGS(iins) & FI_EVAL_B) && (AINS_B_EVAL(iins) | 0x00ffffffffL) == 0x00ffffffffL )
            return 1;
        else if( (AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_A_EVAL(iins) | 0x00ffffffffL) == 0x00ffffffffL )
            return 1;
        else
            return 0;
    default:
        return 0;
    }
}

/*=======================================================
=======================================================*/

extern REG AinsSignExtendsReg(INDEX iins)
{
    switch( AINS_CODE(iins) )
    {
    case I_addl:
        if( AINS_REG_A(iins) == REG_ZERO ) 
            return AINS_REG_B(iins);
        else if( AINS_REG_B(iins) == REG_ZERO ) 
            return AINS_REG_A(iins);
        else
            return INVALID;

    case I_zap:
        if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xf0 )
            return AINS_REG_A(iins);
        else
            return INVALID;
        
    case I_zapnot:
        if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xf )
            return AINS_REG_A(iins);
        else
            return INVALID;
        
    default:
        return INVALID;
    }
}



/*=======================================================
@OPT
Requires: Liveness
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptSignExtensionOne"

extern BOOL OptSignExtensionOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Signextensionone");
    BOOL   change = 0;

    NUMBER tcount = 0;
    NUMBER tcounth = 0;
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    NUMBER count = 0;
    NUMBER counth = 0;
    INT64  weight = 0;
    clock_t time = clock();

    INDEX  ibbl;
    INDEX  iins;
    INDEX  jins;
    REGLIST reglist;
    REG reg;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        /* this has to be backward */
        
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            reg = AinsSignExtendsReg(iins);
            /* disallow literal regs */

            if( reg == REG_LIT ) continue;
            if( !VALID(reg) ) continue;
            
            tcount++;
            if( ABBL_IS_HOT(ibbl) ) tcounth++;

            reglist = REG2BIT(AINS_REG_C(iins));

            
            for( jins=AINS_NEXT(iins); VALID(jins); jins=AINS_NEXT(jins) )
            {
                if( AINS_REG_C(jins) == AINS_REG_C(iins) ) break;
            }
            
            if( !VALID(jins) )
            {
                if( reglist & ABBL_OUT(ibbl)  ) continue;
            }
            
            
            pcount++;
            if( ABBL_IS_HOT(ibbl) ) pcounth++;
            
            for( jins=AINS_NEXT(iins); VALID(jins); jins=AINS_NEXT(jins) )
            {
                if( !(AinsUse(jins) & reglist) ) continue;
                
                if( !AinsDoesOperandSignExtension(jins) ) break;

                if( AINS_REG_C(jins) == AINS_REG_C(iins) )
                {
                    jins = INVALID;
                    break;
                }
            }
            
            if( VALID(jins) ) continue;
            
            count++;
            if( ABBL_IS_HOT(ibbl) )
                counth++;
            weight+=ABBL_COUNT(ibbl);
            
            if( verbose )
            {
                fprintf(fp,"\t[*] converting sign extension into move %d\n",iins);
                PrintBblSuper(stdout,ibbl);
            }
            
            
            AinsMakeLda(iins, reg, AINS_REG_C(iins),0);
            
            if( verbose )
            {
                PrintBblSuper(stdout,ibbl);
            }
            
        }
    }
    
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = count>0;

    InfoStatus(fp,"\t[OptSignextensionone%c] Transformed sign extensions : %d#%d (w: %ld)  Potential: %d#%d Total: %d#%d {%d ms}\n",
            CHANGE(change),count,counth,weight,pcount,pcounth,tcount,tcounth,time);

    return change;
}

/*=======================================================
@OPT
Requires: Liveness
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptSignExtensionTwo"

extern BOOL OptSignExtensionTwo(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Signextensiontwo");
    BOOL   change = 0;

    NUMBER tcount = 0;
    NUMBER tcounth = 0;
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    NUMBER count = 0;
    NUMBER counth = 0;
    INT64  weight = 0;
    clock_t time = clock();

    INDEX  ibbl;
    INDEX  iins;
    INDEX  jins;
    REG reg;

    FOREACH_LIVE_BBL(ibbl)
    {
        /* this has to be backward */
        
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            reg = AinsSignExtendsReg(iins);
            if( !VALID(reg) ) continue;

            tcount++;
            if( ABBL_IS_HOT(ibbl) ) tcounth++;

            jins = AinsFindDefInExtBbl(iins,reg);
            if( !VALID(jins) ) continue;

            pcount++;
            if( ABBL_IS_HOT(ibbl) ) pcounth++;
            
            if( AINS_CODE(iins) == I_zapnot || AINS_CODE(iins) == I_zap )
            {
                switch( AINS_CODE(jins) )
                {
                case I_ldwu:
                case I_ldbu:
                    goto convert;
                case I_zap:
                   if( (AINS_FLAGS(jins) & FI_EVAL_B) && (AINS_B_EVAL(jins) & 0xf0) == 0xf0 )
                       goto convert;
                   break;
                case I_zapnot:
                    if( (AINS_FLAGS(jins) & FI_EVAL_B) && (AINS_B_EVAL(jins) | 0xf) == 0xf )
                       goto convert;
                    break;
                case I_and:
                    if( (AINS_FLAGS(jins) & FI_EVAL_B) && (AINS_B_EVAL(jins) | 0x00ffffffffL) == 0x00ffffffffL )
                        goto convert;
                    else if( (AINS_FLAGS(jins) & FI_EVAL_A) && (AINS_A_EVAL(jins) | 0x00ffffffffL) == 0x00ffffffffL )
                        goto convert;
                    else
                        break;
                }
                
            }
            else if( AINS_CODE(iins) == I_addl )
            {
                switch( AINS_CODE(jins) )
                {
                case I_ldl:
                case I_addl:
                case I_s4addl:
                case I_subl:
                case I_s4subl:
                case I_s8addl:
                case I_s8subl:
                case I_addl_v:
                case I_subl_v:
                case I_mull:
                case I_mull_v:
                    goto convert;
               case I_zap:
                   if( (AINS_FLAGS(jins) & FI_EVAL_B) && (AINS_B_EVAL(jins) & 0xf8) == 0xf8 )
                       goto convert;
                   break;
                case I_zapnot:
                    if( (AINS_FLAGS(jins) & FI_EVAL_B) && (AINS_B_EVAL(jins) | 0x7) == 0x7 )
                       goto convert;
                    break;
                case I_and:
                    if( (AINS_FLAGS(jins) & FI_EVAL_B) && (AINS_B_EVAL(jins) | 0x007fffffffL) == 0x007fffffffL )
                        goto convert;
                    else if( (AINS_FLAGS(jins) & FI_EVAL_A) && (AINS_A_EVAL(jins) | 0x007fffffffL) == 0x007fffffffL )
                        goto convert;
                    else
                        break;
                }
            }
            else
            {
                ERROR(("unexpected sign extension\n"));
            }
            
            continue;
            
        convert:
            count++;
            if( ABBL_IS_HOT(ibbl) )
                counth++;
            weight+=ABBL_COUNT(ibbl);
            
            if( verbose )
            {
                fprintf(fp,"\t[*] converting sign extension into move %d\n",iins);
                PrintBblSuper(stdout,ibbl);
            }
            
            
            AinsMakeLda(iins, reg, AINS_REG_C(iins),0);
            
            if( verbose )
            {
                PrintBblSuper(stdout,ibbl);
            }
            
        }
    }
    
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = count>0;

    InfoStatus(fp,"\t[OptSignextensiontwo%c] Transformed sign extensions : %d#%d (w: %ld)  Potential: %d#%d Total: %d#%d {%d ms}\n",
            CHANGE(change),count,counth,weight,pcount,pcounth,tcount,tcounth,time);

    return change;
}

/*=======================================================
=======================================================*/
      
